##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##

# example of:
# 1) not using elmclient - does all low-level HTTP queries to find the project+component_config and then do OSLC Query
# AND 2) using elmclient to do the same query
# exporting to CSV

# the minimal example has many limitations:
#  only easy for a query which doesn't involve custom attributes/links
#  only gets first page
#  only works for RM
#  ...

import logging

import lxml.etree as ET
import csv

logger = logging.getLogger(__name__)

jazzhost = 'https://jazz.ibm.com:9443'
username = 'ibm'
password = 'ibm'

jtscontext = 'jts'
rmcontext  = 'rm'

projectname = "rm_gc_p1"
componentname = "rm_gc_p1"
configname="rm_gc_p1 Initial Stream"

query = 'dcterms:modified>"" and dcterms:modified<""'
select= "*"
prefix = 'dcterms=<>'

outfile = "dnoq.csv"

prefixes = {
'oslc_rm':
'oslc_config':
'rdf':
'dcterms':
}

#######################################################################################################
#
# low-level version
#

# generic HTTP GET with login if auth is needed
def get_with_optional_login(s,url, *, params=None, headers=None, username="", password="", verify=False):
    '''
    This function tries to perform a GET on url - and checks the result,
    If authentication is needed it does this using provided username/password, and then returns the data
    ONLY WORKS FOR FORM AUTHENTICATION!
    '''
    params = params or {}
    headers = headers or {}
    
    # execute the request - NOTE verify (of SSL cert) is set to False because test server using self-signed cert
    fullurl = 
    
    response = s.get(url,headers=headers, params=params, verify=verify )
    response.raise_for_status()

    # The response is "good" (i.e. 200) but if auth is needed there is a specific header+value
    # (in which case the response is not the data requested, but the data will be returned by authenticating)
    if response.headers.get('X-com-ibm-team-repository-web-auth-msg',"") == 'authrequired':
        print("Auth required")
        # unparse to get the auth URL

        auth_url = f"{host}/jts/j_security_check?j_username={username}&j_password={password}"
        
        response = s.get(auth_url)
        response.raise_for_status()
        # check the auth was successful - is there a JAZZ_AUTH_TOKEN?
        if 'JAZZ_AUTH_TOKEN' not in response.cookies:
            raise Exception ( f"Authentication failed using {username=} {password=}!" )
        print("Auth successful")
    elif # OIDC
    
    elif # FORM prompt (ENI)
    
    else:
    
    return response

# Create the session used for all requests to the server
s = requests.session()

headers = { 'OSLC-Core-Version': '2.0;, 'Accept': 'application/rdf+xml' }

# get rootsservices
response = get_with_optional_login( urllib.urljoin(jazzhost,f"/{rmcontext}/rootservices'),headers=headers, username=username,password=password)
rootservices_x = lxml.fromstring(response.content)

# refer to discovery example here https://jazz.net/forum/questions/266334/dng-oslcfetch-components-from-project-area
cmsp_x = lxml.find( rootservices_x, './/oslc_config:cmServiceProviders' )
cmsp_u = cmsp_x.get('rdf:resource')

# get the compponents entry point
response = get_with_optional_login( cmsp_u,headers=headers, username=username,password=password)
candc_x = lxml.fromstring( response.content )

comps_x = lxml.find( candc_x,'.//oslc:ServiceProvider )
comps_u = comps_x.get( 'rdf:about' )

# get the projects
response = get_with_optional_login( comps_u,headers=headers, username=username,password=password)
projs_x = lxml.fromstring( response.content )

# find the project
project_x = lxml,.find( projs_x, f'.//oslc:CreationFactory/oslc:resourceType[@rdf:resource="http://open-services.net/ns/config#Component"]/../dcterms:title=[text()='{projectname}']/../oslc:creation' )
project_u = projectx.get['rdf:resource']

# find the component
response = get_with_optional_login( project_u,headers=headers, username=username,password=password)
thecomps_x = lxml.fromstring( response.content )
thecomp_x = lxml.find( thecomps_x, f'.//rdf:Description/dcterms:title[text()="{componentname}"]/..' )
thecomp_u = thecomp_x.get('rdf:about')

# find the configuration list pointer
response = get_with_optional_login( thecomp_u,headers=headers, username=username,password=password)
proj_services_x = lxml.fromstring( response.content )
comp_configs_u = lxml.find( proj_services_x, './/oslc_config:configurations' )

# search for the config we want
response = get_with_optional_login( comp_configs_u,headers=headers, username=username,password=password)
aconf_x = lxml.fromstring( response.content )

# search for the config name we want
thisconf_u
for aconf_x in lxml.findall( aconf_x, './/rdfs:member' )
    aconf_u = aconf_x.get( 'rdf;resource' )
    response = get_with_optional_login( aconf_u,headers=headers, username=username,password=password)
    aconf_x = lxml.fromstring( response.content )
    if ( thisconf_x := lxml.find( aconf_x, f'.//oslc_config:Configuration/dcterms:title[text()='{configname}']/..' ) ):
        thisconf_u = thisconf_x.get('rdf:resource')
        break
        
if thisconf_u is None:
    raise Exception( f"Config {configname} not found" )
    
thisconf_x = aconf_x

# setup headers/params for the rest of the operations  -these are all config-specific
params['oslc_config.context'] = thisconf_u
headers[ 'Configuration.context'] = thisconf_u

# retrieve the services.xml for the config
aservices_x = lxml.find( aconf_x, './/oslc:serviceProvider' )
services_u = services_x.get('rdf;resource')

response = get_with_optional_login( services_u,headers=headers, username=username,password=password)
service_x = lxml.fromstring( response.content )

# find the query capability for oslc_rm:Requirement
req_query_x = lxml.find( service_x,'.//oslc:QueryCapability/oslc:resourceType[@rdf:resource="http://open-services.net/ns/rm#Requirement"]/../oslc:queryBase' )
req_query_base_u = req_query.get( 'rdf:resource' )

# build the query URL
# unparse the query capability URL to get any existing parameters, to which we will add the params for this query
url_parts = list(urllib.parse.urlparse(req_query_base_u))
logger.info( f"{url_parts=}" )

# start the parameters from what's in the base url
query_params = dict(urllib.parse.parse_qsl(url_parts[4]))
# override with the query parameters
query_params.update(dict(((k, server.to_binary(v)) for k, v in list(params.items()) if v)))
url_parts[4] = ""
url_parts[5] = ""
# reconstruct just the base scheme:hostname
query_url = urllib.parse.urlunparse(url_parts)
logger.info( f"The full OSLC Query URL is {query_url}" )


# get the query results
response = s.get(query_url,params=query_params, headers=headers )
response_x = lxml.fromstring( response.content )

# process the XML
resultrows = []


# save the results to CSV
print( f"Writing to CSV {outfile}" )

with open( outfile, "w", newline='' ) as csvfile:
    csvwriter = csv.DictWriter(csvfile,fieldnames=sorted(allcolumns))
    csvwriter.writeheader()
    for row in resultrows:
        csvwriter.writerow(row)



#######################################################################################################
#
# elmclient Version
#

import elmclient.server as elmserver
import elmclient.utils as utils



# caching control
# 0=fully cached (but code below specifies queries aren't cached)
# 1=clear cache initially then continue with cache enabled
# 2=clear cache and disable caching
caching = 0

# create our "server" which is how we connect to DOORS Next
# first enable the proxy so if a proxy is running it can monitor the communication with server (this is ignored if proxy isn't running)
elmserver.setupproxy(jazzhost,proxyport=8888)
theserver = elmserver.JazzTeamServer(jazzhost, username, password, verifysslcerts=False, jtsappstring=f"jts:{jtscontext}", appstring='rm', cachingcontrol=caching)

# create the RM application interface
dnapp = theserver.find_app( f"rm:{rmcontext}", ok_to_create=True )


p = dnapp.find_project()
c = dnapp.find_component()
local_config_u = p.find_local_config()
c.set_local_config(local_config_u)
headers = {}

qcbase = c.find_query_capability("oslc_rm:Requirement")

queryurl = .urljoin(qcbase,params)

# now do the direct OSLC query (if the query needed to refer to a specific attribute or enumeration name this would be an awful lot more complicated)
# ONLY get the first page!
response_x = c.app.server.execute_xml_get(queryurl,headers=headers)

# results are in RDF-XML - decode them
resultrows = []

...

print( f"Writing to CSV {outfile}" )

with open( outfile, "w", newline='' ) as csvfile:
    csvwriter = csv.DictWriter(csvfile,fieldnames=sorted(allcolumns))
    csvwriter.writeheader()
    for row in resultrows:
        csvwriter.writerow(row)

print( "Finished" )



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


import logging

from . import rdfxml
from . import utils

logger = logging.getLogger(__name__)

#################################################################################################

class No_Type_System_Mixin():
    def __init__(self,*args,**kwargs):
        self.has_typesystem=False

class Type_System_Mixin():
    def __init__(self,*args,**kwargs):
        self.typesystem_loaded = False
        self.has_typesystem=True
        self.clear_typesystem()

    def clear_typesystem(self):
        self.shapes = {}
        self.properties = {}
        self.enums = {}
        self.values = {}
        self.typesystem_loaded = False
        self._gettypecache = {}

    def textreport(self):

        def quote(s):
            if " " in s:
                return f"'{s}'"
            else:
                return s

        rows = []
        report = ""

        def addtoreport(s, end='\n'):
            nonlocal report
            report += s + end

        reportedproperties = []
        # print a nicely sorted report with shapes at left, then properties (with type, if defined) in that shape, then enumerations in that property
        for shapeuri in sorted(self.shapes.keys(),key=lambda k: self.shapes[k]['name'].lower()):
            rows.append( [f"{quote(self.shapes[shapeuri]['name']):25}"] )
            for propertyuri in sorted(self.shapes[shapeuri]['properties'], key=lambda k: self.properties[k]['name'].lower()):
                reportedproperties.append(propertyuri)
                rows.append( [ "",f"{quote(self.properties[propertyuri]['name'])}"] )
                if self.properties[propertyuri]['altname'] is not None:
                    rows[-1].append( f"{self.properties[propertyuri]['altname']}" )
                else:
                    rows[-1].append("")
                rows[-1].append( f"{rdfxml.uri_to_default_prefixed_tag(propertyuri)}" )
                if self.properties[propertyuri].get('value_type'):
                    rows[-1].append( f"{self.properties[propertyuri]['value_type']}" )
                else:
                     rows[-1].append( "" )
                newrowlen = len(rows[-1])-3
                # add enums as additional rows
                for enum_uri in sorted(self.properties[propertyuri]['enums'],key=lambda k:self.enums[k]['name'].lower()):
                    eid = self.enums[enum_uri].get('id') or enum_uri
                    rows.append( [""]*newrowlen+[f"{quote(self.enums[enum_uri]['name'])}",eid,enum_uri ] )
                    logger.info( f"appended for enum {rows[-1]}" )
        if len(rows)>0:
            addtoreport( "<h2>Shapes<h2>\n" )
            report += utils.print_in_html( rows,['Shape','Property Name','Property label','URI'] )

        # now report properties without shape
        rows = []
        for propertyuri in sorted(self.properties, key=lambda k: self.properties[k]['name'].lower()):
            if propertyuri not in reportedproperties:
                rows.append( [ f"{quote(self.properties[propertyuri]['name'])}" ] )
                if self.properties[propertyuri]['altname'] is not None:
                    rows[-1].append( f"{self.properties[propertyuri]['altname']}" )
                else:
                    rows[-1].append("")
                rows[-1].append( f"{rdfxml.uri_to_default_prefixed_tag(propertyuri)}" )
#                    addtoreport( f"{INDENT}{propertyuri}", end="" )
                if self.properties[propertyuri].get('value_type'):
                    rows[-1].append( f"{self.properties[propertyuri]['value_type']}" )
                else:
                    rows[-1].append( "" )
                newrowlen = len(rows[-1])-3
                # add enums as additional rows
                for enum_uri in sorted(self.properties[propertyuri]['enums'],key=lambda k:self.enums[k]['name'].lower()):
                    eid = self.enums[enum_uri].get('id') or enum_uri
                    rows.append( [""]*newrowlen+[f"{quote(self.enums[enum_uri]['name'])}",eid,enum_uri ] )
                    logger.info( f"appended for enum {rows[-1]}" )

        if len(rows)>0:
            addtoreport( "<h2>Properties with no shape</h2>\n" )
            report += utils.print_in_html( rows,['Shape','Property Name','Property label','URI'] )

        return report


    # normalise results to either a URI or if a tag expand it, or the name
    def normalise_uri( self, uri, exception_if_name=False ):
        if uri is None:
            result = None
        elif uri.startswith( 'http://') or uri.startswith( 'https://'):
            result = uri
        elif ':' in uri:
            result = rdfxml.tag_to_uri( uri )
            logger.info( f"tag_to_uri {uri=} {result=}" )
        else:
            raise Exception( f"Expecting a uri but this doesn't look like a URI {uri}" )
        return result

    def is_known_shape_uri(self,shape_uri ):
        logger.info( f"is_known_shape_uri {shape_uri=}" )
        shape_uri = self.normalise_uri( shape_uri )
        result = self.shapes.get(shape_uri) is not None
        logger.info( f"is_known_shape_uri {shape_uri=} returning {result=}" )
        return result

    def register_shape( self, shape_name, shape_uri ):
        logger.info( f"register_shape {shape_name=} {shape_uri=}" )
        shape_uri = self.normalise_uri( shape_uri)
        if shape_uri in self.shapes:
            raise Exception( f"Shape {shape_uri} already defined!" )
        # add the URI as the main registration for the shape
        self.shapes[shape_uri] = {'name':shape_name,'shape':shape_uri,'properties':[]}
        self.loaded = True

    def get_shape_uri( self, shape_name ):
        logger.info( f"get_shape_uri {shape_name=}" )
        shapes = [k for k,v in self.shapes.items()  if v['name']==shape_name ]
        if len(shapes)==1:
            result = shapes[0]
        else:
            result = None
        return result

    def get_shape_name( self, shape_uri ):
        shape_uri = self.normalise_uri( shape_uri)
        result = self.shapes.get(shape_uri)
        return result

    def is_known_property_uri( self, property_uri, *, shape_uri=None, raiseifnotfound=True ):
        logger.info( f"is_known_property_uri {property_uri=} {shape_uri=}" )
        property_uri = self.normalise_uri( property_uri )
        shape_uri = self.normalise_uri( shape_uri )
        if property_uri in self.properties:
            if self.properties[property_uri]['shape']==shape_uri:
                result = True
            else:
                if raiseifnotfound:
                    raise Exception( f"Property {property_uri} not registered with shape {shape_uri}" )
                result = False
        else:
            result = False
        logger.info( f"is_known_property_uri {property_uri=} {shape_uri=} returning {result=}" )
        return result

    def register_property( self, property_name, property_uri, *, property_value_type=None, shape_uri=None, altname = None, do_not_overwrite=True, property_definition_uri=None ):
        logger.info( f"register_property {property_name=} {property_uri=} {shape_uri=}" )
        property_uri = self.normalise_uri( property_uri )
        shape_uri = self.normalise_uri( shape_uri )
        if not do_not_overwrite or property_uri not in self.properties:
            self.properties[property_uri] = {'name': property_name, 'shape': shape_uri, 'enums': [], 'value_type': property_value_type, 'altname':altname}
        if altname and property_definition_uri and ( not do_not_overwrite or property_definition_uri not in self.properties):
            self.properties[property_definition_uri] = {'name': altname, 'shape': shape_uri, 'enums': [], 'value_type': property_value_type, 'altname':None}
            self.properties[rdfxml.uri_to_default_prefixed_tag(property_definition_uri)] = {'name': altname, 'shape': shape_uri, 'enums': [], 'value_type': property_value_type, 'altname':None}
        if shape_uri is not None:
            self.shapes[shape_uri]['properties'].append(property_uri)
        self.loaded = True

    def get_property_uri( self, property_name, *, shape_uri=None ):
        logger.info( f"get_property_uri {property_name=} {shape_uri=}" )
        shape_uri = self.normalise_uri( shape_uri )
        properties = [k for k,v in self.properties.items() if v['name']==property_name and v['shape']==shape_uri]
        if len(properties)==1:
            result = properties[0]
        else:
            # try using altname
            altproperties = [k for k,v in self.properties.items() if v['altname']==property_name and v['shape']==shape_uri]
            if len(altproperties)==1:
                result = altproperties[0]
                logger.info( f"Property {property_name} found using altname" )
            else:
                if len(altproperties)>1:
                    altnames = [self.properties[k]['altname'] for k in properties]
                    raise Exception( f"Property {property_name} is ambiguous - maybe use the altname - {altnames}" )
                else:
                    # try for a property ignoring the shape - as long as all the ones with the name have the same URI after normalising to a uri if tag/prefix present
                    properties = [k for k,v in self.properties.items() if v['name']==property_name]
                    if len(properties)==1 or (len(properties)>1 and all([rdfxml.tag_to_uri(k)==rdfxml.tag_to_uri(properties[0]) for k in properties[1:]]) ):
                        result = properties[0]
                    else:
                        result = None
        logger.info( f"get_property_uri {property_name=} {shape_uri=} returning {result=}" )
        return result

    def get_property_name( self, property_uri, shapeuri=None ):
        logger.info( f"get_property_name {property_uri=} {shape_uri=}" )
        property_uri = self.normalise_uri( property_uri )
        result = self.properties.get(property_uri)
        return result

    def is_known_enum_uri( self, enum_uri ):
        enum_uri = self.normalise_uri( enum_uri )
        result = self.enums.get(enum_uri)
        logger.info( f"is_known_enum_uri {enum_uri=} returning {result=}" )
        return result

    def register_enum( self, enum_name, enum_uri, property_uri, *, id=None ):
        logger.info( f"register_enum {enum_name=} {enum_uri=} {property_uri=} {id=}" )
        # add the enum  to the property
        enum_uri = self.normalise_uri( enum_uri )
        property_uri = self.normalise_uri( property_uri )
        self.enums[enum_uri] = {'name': enum_name, 'id':id, 'property': property_uri}
        self.properties[property_uri]['enums'].append(enum_uri)
        self.loaded = True

    def get_enum_uri(self, enum_name, property_uri):
        property_uri = self.normalise_uri( property_uri )
        result = None
        for enumuri in self.properties[property_uri]['enums']:
            if self.enums[enumuri]['name']==enum_name:
                result = enumuri
                break
        return result

    def get_enum_name( self, enum_uri ):
        property_uri = self.normalise_uri( property_uri )
        return self.enums[enum_uri]['name']

    def get_enum_id( self, enum_name, property_uri ):
        logger.info( f"get_enum_id {enum_name=} {property_uri=}" )
        property_uri = self.normalise_uri( property_uri )
        result = None
        logger.info( f"{self.properties[property_uri]=}" )
        logger.info( f"{self.properties[property_uri]['enums']=}" )
        for enum_uri in self.properties[property_uri]['enums']:
            if self.enums[enum_uri]['name']==enum_name:
                result = self.enums[enum_uri]['id'] or enum_uri
                break
        logger.info( f"get_enum_id {enum_name=} {property_uri=} {result=}" )
        return result

    # generic uri/name

    def is_known_uri( self, uri ):
        logger.debug( f"iku {uri}" )
        uri = self.normalise_uri( uri )
        result =  ( self.shapes.get(uri) or self.properties.get(uri) or self.enums.get(uri) or self.values.get(uri) ) is not None
        logger.info( f"is_known_uri {uri=} returning {result=} s={self.shapes.get(uri)} p={self.properties.get(uri)} e={self.enums.get(uri)} v={self.values.get(uri)}" )
        return result

    def register_name( self, name, uri ):
        uri = self.normalise_uri( uri )
        self.values[uri]={'name': name }
        self.loaded = True

    def get_uri_name( self, uri ):
        uri = self.normalise_uri( uri )
        result = self.shapes.get(uri) or self.properties.get(uri) or self.enums.get(uri) or self.values.get(uri)
        if result is not None:
            result = result['name']
        logger.info( f"get_uri_name {uri=} returning {result=}" )
        return result

    def get_name_uri( self, name ):
        result = self.get_shape_uri(name) or self.get_property_uri(name) or self.get_enum_uri(name) or self.get_value_uri(name)
        return result



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


#################################################################################################

# ENI/RELM application

#
# This is a skeleton implementation for ENI/RELM - just enough to support finding a project
#

#################################################################################################

import logging

import requests
import lxml.etree as ET
import tqdm

from . import _app
from . import _project
from . import _typesystem
from . import oslcqueryapi
from . import rdfxml
from . import server
from . import utils

logger = logging.getLogger(__name__)

#################################################################################################

class _RELMProject(_project._Project):
    def __init__(self, name, project_uri, app, is_optin=False, singlemode=False,defaultinit=False):
        super().__init__(name, project_uri, app, is_optin,singlemode, defaultinit=defaultinit)
        self.hooks = []
        self._components = None  # will become a dict keyed on component uri
        self._configurations = None # keyed on the config name
        self.default_query_resource = 'oslc_config:Configuration'


#################################################################################################

#@utils.mixinomatic
class RELMApp(_app._App):
    domain = 'relm'
    project_class = _RELMProject
    supports_configs = False
    supports_components = False
    supports_reportable_rest = False

    relprefixes = (
        )

#    identifier_name = 'Short ID'
#    identifier_uri = 'Identifier'

    def __init__(self, server, contextroot, jts=None):
        super().__init__(server, contextroot, jts=jts)

        self.rootservices_xml = self.execute_get_xml( self.reluri('rootservices'), intent="Retrieve RELM/ENI root services" )
#        self.serviceproviders = 'gc:globalConfigServiceProviders'
#        self.default_query_resource = 'oslc_config:Configuration'
        # register some app-specific namespaces
        for prefix,reluri in self.relprefixes:
            rdfxml.addprefix(prefix,self.baseurl+reluri)
        self.hooks = []

    def _get_headers(self, headers=None):
        result = super()._get_headers()
        result['net.jazz.jfs.owning-context'] = self.baseurl
        if headers:
            result.update(headers)
        return result




##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


import logging

import lxml.etree as ET

logger = logging.getLogger(__name__)

# Some well-known and used RDF/XML prefixes
RDF_DEFAULT_PREFIX = {
    'acc':          'http://open-services.net/ns/core/acc#',  # added for GCM
    'acp':          'http://jazz.net/ns/acp#',
    'config_ext':   'http://jazz.net/ns/config_ext#',
    'dc':           'http://purl.org/dc/elements/1.1/',
    'dcterms':      'http://purl.org/dc/terms/',
    'dng_reqif':    'http://jazz.net/ns/rm/dng/reqif#',
    'dng_task':     'http://jazz.net/ns/rm/dng/task#',      # only for task tracker
    'foaf':         'http://xmlns.com/foaf/0.1/',
    'gc':           'http://jazz.net/ns/globalconfig#',
    'jazz_rm':      'http://jazz.net/ns/rm#',
    'jfs':          'http://jazz.net/xmlns/prod/jazz/jfs/1.0/',
    'jp':           'http://jazz.net/xmlns/prod/jazz/process/1.0/',
    'jp06':         'http://jazz.net/xmlns/prod/jazz/process/0.6/',
    'ldp':          'http://www.w3.org/ns/ldp#',
    'ns':           'http://com.ibm.rdm/navigation#',
    'oslc':         'http://open-services.net/ns/core#',
    'oslc_cm':      'http://open-services.net/xmlns/cm/1.0/',
    'oslc_cm1':     'http://open-services.net/ns/cm#',
    'oslc_cm_10':   'http://open-services.net/xmlns/cm/1.0/cm#',
    'oslc_config':  'http://open-services.net/ns/config#',
    'oslc_qm':      'http://open-services.net/ns/qm#',
    'oslc_qm_10':   'http://open-services.net/xmlns/qm/1.0/',
    'oslc_rm':      'http://open-services.net/ns/rm#',
    'oslc_rm_10':   'http://open-services.net/xmlns/rm/1.0/',
    'owl':          'http://www.w3.org/2002/07/owl#',
    'process':      'http://jazz.net/ns/process#',
    'public_rm_10': 'http://www.ibm.com/xmlns/rm/public/1.0/',
    'prov':         'http://www.w3.org/ns/prov#',  # added for GCM
    'qm_rqm':       "http://jazz.net/ns/qm/rqm#",
    'rdf':          'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'rdfs':         'http://www.w3.org/2000/01/rdf-schema#',
    'rdm_types':    'http://www.ibm.com/xmlns/rdm/types/',
    'rm':           'http://www.ibm.com/xmlns/rdm/rdf/',
    'rm_ds':        'http://jazz.net/xmlns/alm/rm/datasource/v0.1', # For RR
    'rm_modules':   'http://jazz.net/ns/rm/dng/module#',
    'rm_config':    'http://jazz.net/ns/rm/dng/config#',
    'rm_nav':       'http://jazz.net/ns/rm/navigation#',
    'rm_text':      'http://jazz.net/xmlns/alm/rm/text/v0.1', # for RR
    'rm_view':      'http://jazz.net/ns/rm/dng/view#',
    'rqm':          'http://jazz.net/xmlns/prod/jazz/rqm/qm/1.0/',
    'rrm':          'http://www.ibm.com/xmlns/rrm/1.0/', # For RR
    'rtc_cm':       "http://jazz.net/xmlns/prod/jazz/rtc/cm/1.0/",
    'xhtml':        'http://www.w3.org/1999/xhtml',
    'xml':          'http://www.w3.org/XML/1998/namespace',
    'xsd':          'http://www.w3.org/2001/XMLSchema#',
    'xs':          'http://www.w3.org/2001/XMLSchema'
}

# Register prefixes to XML system
for prefix,uri in list(RDF_DEFAULT_PREFIX.items()):
    ET.register_namespace(prefix, uri)

def addprefix(prefix,uri,prefix_map=RDF_DEFAULT_PREFIX):
    logger.info( f"Adding prefix {prefix=} {uri=}" )
    if  ( not uri.startswith('http://') and not uri.startswith('https://') ) or (not uri.endswith('#') and not uri.endswith('/')):
        raise Exception( f"BAD URI {prefix=} {uri}" )
    uses = [k for k,v in prefix_map.items() if v==uri]
    if uses:
        logger.info( f"Warning prefix {prefix} uri {uri} is already used for prefex {uses[0]}" )
    prefix_map[prefix]=uri
    ET.register_namespace(prefix, uri)

# find single element matching element_xpath, and checks them for any subelement matching condition_xpath and if condition_value is specified then also matching it
# exception if more than 1 match!
def xml_find_element(xml, element_xpath, condition_xpath=None, condition_value=None, strip=True, prefix_map=RDF_DEFAULT_PREFIX):
    results = xml_find_elements(xml, element_xpath, condition_xpath=condition_xpath, condition_value=condition_value, strip=strip, prefix_map=prefix_map)
    if len(results) > 0:
        if len(results) > 1:
            raise Exception(f"Multiple results when single result expected! {results}")
        return results[0]
    return None

# find all elements matching element_xpath, and checks them for any subelement matching condition_xpath and if condition_value is specified then also matching it

# condition_xpath is xpath with additional features:
# ending in /#text compares the .text with condition_value - selects element if it matches
# ending in /@attrib compares value of attrib with condition_value - selects element if it matches
# default is to match .text with condition_value
# or if no condition_value is specified then the element is added to results

def xml_find_elements(xml, element_xpath, condition_xpath=None, condition_value=None, strip=True,
                      prefix_map=RDF_DEFAULT_PREFIX):
    elements = xml.findall(element_xpath, prefix_map)
    def eq(left, right):
        if strip and left and right:
            return left.strip() == right.strip()
        else:
            return left == right

    results = []
    if condition_xpath:
        # get the last special part of the condition_xpath, if there is one
        last_slash_pos = condition_xpath.rfind('/')
        cond_xp = condition_xpath
        cond_spec = None
        if last_slash_pos >=0:
            last_part = condition_xpath[last_slash_pos + 1:]
            if last_part == '#text' or last_part.startswith('@'):
                # the special last parts shorten the condition_xpath
                cond_xp = condition_xpath[:last_slash_pos]
                cond_spec = last_part

        for e in elements:
            # check ALL matching subelements (used just to use e.find so only tried matching first subelement, which failed e.g. when looking for QueryCapability)
            subels = e.findall(cond_xp, prefix_map)
            for x in subels:
                if cond_spec and condition_value:
                    if cond_spec == '#text':
                        if eq(x.text, condition_value):
                            results.append(e)
                            # no need to check any more subelements
                            break
                    elif cond_spec.startswith('@'):
                        if eq(x.get(uri_to_tag(cond_spec[1:])), condition_value):
                            results.append(e)
                            # no need to check any more subelements
                            break
                    else:
                        raise Exception( f"unexpected special value {cond_spec} in xml_find_elements()" )
                else:
                    if condition_value:
                        if eq(str(x.text), str(condition_value)):
                            results.append(e)
                            # no need to check any more subelements
                            break
                    else:
                        results.append(e)
                        # no need to check any more subelements
                        break
    else:
        results = list(elements)

    return results

# finds first element using xpath and returns (in order of preference) its rdf attrib, rdf:resource, rdf:about, or text
def xmlrdf_get_resource_uri(xml, xpath=None, prefix_map=RDF_DEFAULT_PREFIX,attrib=None):
    if xml is None:
        return None
    r = xml.find(xpath, prefix_map) if xpath else xml
    if r is not None:
        if attrib is not None:
            result = r.get(uri_to_tag(attrib, prefix_map=prefix_map))
            return result
        result = r.get(uri_to_tag('rdf:resource', prefix_map=prefix_map))
        if result:
            return result
        result = r.get(uri_to_tag('rdf:about', prefix_map=prefix_map))
        if result:
            return result
        result = r.text
        if result:
            return result
    return None


# finds first element using xpath and return its text value
def xmlrdf_get_resource_text(xml, xpath, prefix_map=RDF_DEFAULT_PREFIX):
    r = xml.find(xpath, prefix_map)
    if r is not None:
        result = r.text
        return result
    return None


# The term "tag" usually refers to an ElementTree-style tag "{ns}id"
# The term "prefixed tag" refers to an XML namespaced tag like "ns:id"

# return the full uri for a tag like rdf:a or {rdf}a
def tag_to_uri(tag, prefix_map=RDF_DEFAULT_PREFIX,noexception=False):
    if tag is None:
        return None
    pos_colon = tag.find(':')
    if tag.find('/') < 0 and pos_colon >= 0:
        prefix = tag[:pos_colon]
        if prefix not in prefix_map:
            if not noexception:
                raise Exception("Prefix is not resolved: %s" % tag)
            else:
                return tag
        return prefix_map[prefix] + tag[pos_colon + 1:]
    # as the tag contains a / or does not contain :, simply remove the { } and return the resulting URI
    return tag.replace('{', '').replace('}', '')


# return the tag equivalent of a uri - tag is in form e.g. {rdf}a
def uri_to_tag(uri, prefix_map=RDF_DEFAULT_PREFIX):
    if uri is None:
        return None
    pos_colon = uri.find(':')
    if uri.find('/') < 0 and pos_colon >= 0:
        prefix = uri[:pos_colon]
        if prefix not in prefix_map:
            raise Exception("Prefix is not resolved: %s" % uri)
        return '{' + prefix_map[prefix] + '}' + uri[pos_colon + 1:]
    pos = max(uri.rfind('#'), uri.rfind('/'))
    if pos == -1:
        return uri
    return '{' + uri[:pos + 1] + '}' + uri[pos + 1:]


# convert tag to XML-style prefixed tag
def tag_to_prefix(tag):
    if tag is None:
        return None
    return tag[tag.find('{') + 1: tag.find('}')]

# convert uri to a prefixed tag, creating a new prefix if one doesn't already exist
# of if uri is already a prefixed tag, make sure prefix is put in uri_to_prefix_map
# uri_to_prefix_map is updated with the new prefix (this is the project-specific prefixes)
# default_map is also updated (this is all the application-wide prefixes)
def uri_to_prefixed_tag(uri, uri_to_prefix_map=None, default_map=None,oktocreate=True,noexception=False):
    logger.debug( f"uri_to_prefixed_tag {uri=}" )
    if uri_to_prefix_map is None: # this allows caller to pass an empty dictionary which will be updated (the usual 'or' test doesn't allow returning!)
        uri_to_prefix_map = {}
    if default_map is None:
        default_map = RDF_DEFAULT_PREFIX
    pos = max(uri.rfind('#'), uri.rfind('/'))
    if pos == -1:
        # this could already be prefixed
        if ':' in uri:
            if not uri.startswith('http:') and not uri.startswith('https:'):
                prefix,rest = uri.split(':',1)
                if ' ' in prefix:
                    return uri
                if prefix not in uri_to_prefix_map:
                    if prefix not in default_map:
                        if not noexception:
                            raise Exception( "Unknown prefix {prefix}" )
                        else:
                            return uri
                    # copy from default map
                    logger.debug( "Copying prefix from default {uri} {prefix} {default_map[prefix]}" )
                    uri_to_prefix_map[default_map[prefix]]=prefix
        else:
            if not noexception:
                raise Exception( f"Unknown uri {uri}" )

        return uri
    if not uri.startswith('http:') and not uri.startswith('https:'):
        return uri
    ns_uri = uri[:pos + 1]
    logger.debug( f"{ns_uri=}" )
    if ns_uri in uri_to_prefix_map:
        prefix = uri_to_prefix_map[ns_uri]
        logger.debug( f"found prefix {prefix=} {ns_uri=}" )
    else:
        if ns_uri in list(default_map.values()):
            # use the existing prefix
            prefix = [k for (k, v) in list(default_map.items()) if v == ns_uri][0]
            logger.debug( f"found prefix in default map {prefix=} {ns_uri=}" )
            prefixok = True
        else:
            logger.debug( f"Creating new prefix {ns_uri=}" )
            prefixok=False
            if oktocreate:
                # create a new prefix
                for i in range(1000):
                    prefix = 'rp' + str(i)
                    if prefix not in list(uri_to_prefix_map.values()) and prefix not in default_map.keys():
                        prefixok = True
                        logger.debug( f"New prefix {prefix}" )
                        break

        if prefixok:
            uri_to_prefix_map[ns_uri] = prefix
            default_map[prefix] = ns_uri
            logger.debug( f"Defined new prefix {prefix} {ns_uri}" )
        elif not noexception:
            raise Exception("Not allowed to create a new prefix for URI: %s" % ns_uri)
        else:
            logger.debug( f"FAILED PREFIX CREATION for {uri}" )
            return uri

    return prefix + ':' + uri[pos + 1:]


# if the URI matches an existing prefix mapping (ending with # or /) then apply it and return prefixed tag, otherwise don't - return the value
def uri_to_default_prefixed_tag(uri, default_map=RDF_DEFAULT_PREFIX):
    if not uri.startswith('http:') and not uri.startswith('https:'):
        return uri
    pos = max(uri.rfind('#'), uri.rfind('/'))
    if pos == -1:
        return uri
    ns_uri = uri[:pos + 1]
    if ns_uri in list(RDF_DEFAULT_PREFIX.values()):
        prefix = [k for (k, v) in list(default_map.items()) if v == ns_uri][0]
        return prefix + ':' + uri[pos + 1:]
    return uri


def remove_tag(s):
    if s.startswith('{'):
        s = s[s.find('}') + 1:]
    return s




##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


import logging

import anytree
import lxml.etree as ET
import requests
import tqdm

from . import _app
from . import _project
from . import _typesystem
from . import oslcqueryapi
from . import rdfxml
from . import server
from . import utils

#################################################################################################

logger = logging.getLogger(__name__)

#################################################################################################

class _CCMProject(_project._Project):
    def __init__(self, name, project_uri, app, is_optin,singlemode):
        super().__init__(name, project_uri, app, is_optin,singlemode)
        self.default_query_resource = 'oslc_cm1:ChangeRequest'

    # for CCM, load the typesystem using all resourceshapes in the services xml
    def _load_types(self,force=False):
        if self.typesystem_loaded and not force:
            return
        self.clear_typesystem()

        sx = self.get_services_xml()

        if sx:
            shapes_to_load = rdfxml.xml_find_elements(sx, './/oslc:resourceShape')

            pbar = tqdm.tqdm(initial=0, total=len(shapes_to_load),smoothing=1,unit=" results",desc="Loading EWM shapes")

            for el in shapes_to_load:
                self._load_type_from_resource_shape(el)
                pbar.update(1)

            pbar.close()

            return
        self.typesystem_loaded = True
        return None

    # pick all the attributes from a resource shape definition
    # and for enumerated attributes get all the enumeration values
    def _load_type_from_resource_shape(self, el, supershape=None):
        uri = el.get("{%s}resource" % rdfxml.RDF_DEFAULT_PREFIX["rdf"])
        logger.info( f"_load_type_from_resource_shape {el=} {uri=}" )
        try:
            if not self.is_known_shape_uri(uri):
                logger.info( f"Starting shape {uri} =======================================" )
                logger.debug( f"Getting {uri}" )
                shapedef = self._get_typeuri_rdf(uri)
                name = rdfxml.xml_find_element(shapedef, f".//rdf:Description[@rdf:about='{uri}']/dcterms:title").text
                self.register_shape( name, uri )
                logger.info( f"Opening shape {name} {uri}" )
            else:
                logger.info( "ENDED" )
                return
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                logger.debug("Failed because type not found - ignoring!", e)
                return
            else:
                raise
                # find the   <rdf:Description rdf:about="https://jazz.net/jazz03/oslc/context/_s_M8MMFIEdumDPUtna_vLA/shapes/workitems/defect">
        # go through all contained    <oslc:property rdf:resource="https://jazz.net/jazz03/oslc/context/_s_M8MMFIEdumDPUtna_vLA/shapes/workitems/defect/property/howfound"/>
        # for each find the   <rdf:Description rdf:about="https://jazz.net/jazz03/oslc/context/_s_M8MMFIEdumDPUtna_vLA/shapes/workitems/defect/property/howfound">
        # get its nested <dcterms:title rdf:parseType="Literal">How Found</dcterms:title>
        # save as defn.property
        # Either
        #   check for contained     <oslc:allowedValues rdf:resource="https://jazz.net/jazz03/oslc/context/_s_M8MMFIEdumDPUtna_vLA/shapes/workitems/defect/property/howfound/allowedValues"/>
        #   find   <rdf:Description rdf:about="https://jazz.net/jazz03/oslc/context/_s_M8MMFIEdumDPUtna_vLA/shapes/workitems/defect/property/howfound/allowedValues">
        #   each contained enum value <oslc:allowedValue rdf:resource="https://jazz.net/jazz03/oslc/enumerations/_s_M8MMFIEdumDPUtna_vLA/howfound/howfound.literal.l20"/>
        #   use the last \w+ of the rdf:resource URL as the name of the enum value
        # OR
        #     <oslc:range rdf:resource="https://jazz.net/jazz03/oslc/enumerations/_s_M8MMFIEdumDPUtna_vLA/OS"/>
        #     lists the enumeration values somehow
        # leave enumeration values to be resolved only when the type is used so we don't spend forever retrieving unused definitions
        n = 0
        # find the list of attributes
        thisshapedef = rdfxml.xml_find_element( shapedef,f'.//rdf:Description[@rdf:about="{uri}"]' )
        if thisshapedef is None:
            raise Exception( f"Shape definition for {uri} not found!" )

        title = rdfxml.xmlrdf_get_resource_text(thisshapedef,'./dcterms:title')

        # scan the attributes
        for propel in rdfxml.xml_find_elements( thisshapedef,'./oslc:property' ):
            propuri = rdfxml.xmlrdf_get_resource_uri(propel)
            real_propel = rdfxml.xml_find_element(shapedef, f'.//rdf:Description[@rdf:about="{propuri}"]')
            property_title = rdfxml.xml_find_element(real_propel, './dcterms:title').text
#            property_uri = rdfxml.xmlrdf_get_resource_uri(real_propel,'./oslc:valueShape')
            proptype = rdfxml.xmlrdf_get_resource_uri(real_propel,'./oslc:valueType' )
            property_definition_uri = rdfxml.xmlrdf_get_resource_uri(real_propel,'./oslc:propertyDefinition' )

            if self.is_known_property_uri( propuri ):
                logger.debug( f"ALREADY KNOWN" )
                continue

            # EWM seems to not infrequently have repeated identical property titles on a shape, so let's create an alternative name that can be used to disambiguate
            # (at least these don't have duplicates AFAICT)
            altname  = propuri[propuri.rfind("/")+1:]
            if len(altname)==0 or altname==property_title:
                altname = None

            logger.info( f"Defining property {title}.{property_title} {altname=} {propuri=} +++++++++++++++++++++++++++++++++++++++" )
            self.register_property(property_title,propuri, shape_uri=uri, altname=altname,property_definition_uri=property_definition_uri)

            allowedvaluesu = rdfxml.xmlrdf_get_resource_uri(real_propel, ".//oslc:allowedValues" )
            if allowedvaluesu is not None:
                # get the enumeration definitions
                thisenumrangex = rdfxml.xml_find_element( shapedef,f'.//rdf:Description[@rdf:about="{allowedvaluesu}"]' )
                # retrieve each definition
                nvals = 0
                for enumvalx in rdfxml.xml_find_elements( thisenumrangex,'.//oslc:allowedValue'):
                    logger.debug( f"{enumvalx=}" )
                    enum_uri = rdfxml.xmlrdf_get_resource_uri(enumvalx)
                    logger.debug( f"{enum_uri=}" )
                    nvals += 1
                    if not self.is_known_enum_uri( enum_uri ):
                        # retrieve it and save the enumeration name and uri in types cache
                        try:
                            enumx = self._get_typeuri_rdf(enum_uri)
                            enum_value_name = rdfxml.xmlrdf_get_resource_uri(enumx, './/rdf:Description/dcterms:title')
                            enum_id = rdfxml.xmlrdf_get_resource_uri(enumx, './/rdf:Description/dcterms:identifier' )
                        except requests.HTTPError as e:
                            if e.response.status_code in [403,404,406]:
                                logger.debug( f"Type {uri} doesn't exist!" )
                            else:
                                raise
                            logger.info( f"No title for {enum_uri}" )
                            enum_id=None
                            enum_value_name = None
                        if enum_value_name is None:
                            enum_value_name = enum_uri

                        if enumx is None:
                            logger.info( "Enumx is None - skipping the type" )
                            continue

                        logger.info( f"defining enum value {enum_value_name=} {enum_id=} {enum_uri=}" )
                        if not self.app.server.jts.is_user_uri(uri):
                            self.register_enum( enum_value_name, enum_uri, id=enum_id, property_uri=propuri )
                        else:
                            logger.debug( f"Not registering enum value for user {uri}" )
                        # (only) the various work item types have a 'category' - this can be used to filter for them on property rtc_cm:type
                        category = rdfxml.xmlrdf_get_resource_text(enumx, './/rdf:Description/rtc_cm:category' )
                        if category is not None:
                            # register this shape as an rtc_cm:type enum
                            logger.info( f"Defining category rtc_cm:type {enum_value_name} {enum_id} {category} {enum_uri}" )
                            # ensure the rtc_cm:type property is defined (but don't overwrite existing definition)
                            self.register_property( 'rtc_cm:type', 'rtc_cm:type', do_not_overwrite=True )
                            # add the shape to it using the shape's URI as an enum URI
                            # NOTE the use of the id - this id is used when comparing values with rtc_cm_type to see workaround https://jazz.net/forum/questions/86619/oslc-20-query-with-oslcwhere-parameter-dctermstype-returns-400-unknown-attribute-id
                            self.register_enum( enum_value_name, enum_uri, 'rtc_cm:type',id=enum_id )

                if nvals==0:
                    raise Exception( f"Enumeration {propuri} with no values loaded" )
        return n


    # for OSLC query, given a type URI, return its name
    # ccm-specific resolution
    def app_resolve_uri_to_name(self, uri):
        if self.is_resource_uri(uri):
            result = self.resource_id_from_uri(uri)
        elif self.is_type_uri(uri):
            result = self.type_name_from_uri(uri)
        else:
            result = None
        return result

    def get_missing_uri_title( self,uri):
        id = None
        if uri.startswith( self.app.baseurl ):
            try:
                resource_xml = self.execute_get_rdf_xml(reluri=uri, intent="Retrieve type definition to get its title" )
                id = rdfxml.xmlrdf_get_resource_text(resource_xml, ".//dcterms:title")
            except ET.XMLSyntaxError as e:
                logger.debug( f"Type {uri} doesn't exist (not XML)!" )
            except requests.HTTPError as e:
                if e.response.status_code==404 or e.response.status_code==406:
                    logger.debug( f"Type {uri} doesn't exist!" )
                else:
                    raise
        if id is None and ( uri.startswith( "http://" ) or uri.startswith( "https://" ) ):
            uri1 = rdfxml.uri_to_prefixed_tag(uri)
            logger.debug( f"Returning the raw URI {uri} so changed it to prefixed {uri1}" )
            id = uri1

        logger.debug( f"gmut {id=}" )

        return id

    # for OSLC query, given a type URI, return the type name
    def type_name_from_uri(self, uri):
        if self.is_type_uri(uri):
            try:
                resource_xml = self.execute_get_rdf_xml( reluri=uri, intent="Retrieve type definition to get its identifier" )
                id = rdfxml.xmlrdf_get_resource_text(resource_xml, ".//dcterms:identifier")
            except requests.HTTPError as e:
                if e.response.status_code==404 or e.response.status_code==406:
                    logger.debug( f"Type {uri} doesn't exist!" )
                else:
                    raise
                id = uri
            return id
        raise Exception(f"Bad type uri {uri}")

    # for OSLC query, given a resource URI, return identifier - for CCM this is the last part of the URI
    def resource_id_from_uri(self, uri):
        if self.is_resource_uri(uri):
            id = uri[uri.rfind("/")+1:]
            return id
        raise Exception(f"Bad resource uri {uri}")

    def is_resource_uri(self, uri):
        if uri and uri.startswith(self.app.baseurl) and '/resources/' in uri:
            return True
        return False

    def is_type_uri(self, uri):
        if uri and uri.startswith(self.app.baseurl) and '/oslc/context/' in uri:
            return True
        return False

#################################################################################################

@utils.mixinomatic
class _CCMApp(_app._App, _typesystem.No_Type_System_Mixin):
    domain = 'ccm'
    project_class = _CCMProject
    supports_configs = False
    supports_components = False
    reportablerestbase='rpt/repository'
    supports_reportable_rest = True
    reportable_rest_status = "Application supports Reportable REST but represt for ccm not fully tested/working"
    artifact_formats = [ # For RR
        'foundation'
        ,'scm'
        ,'build'
        ,'apt'
        ,'workitem'
        ]
    identifier_name = 'id'
    identifier_uri = 'dcterms:identifier'
    # these are schema xpath paths of known queryable and unqueryable paths
    _rr_queryable = [
        'workitem/projectArea/name',
        'workitem/workItem/id',
        'workitem/workItem/type/id',
        'workitem/workItem/target/name',
        'workitem/workItem/tags',
        'workitem/workItem/state/id',
        'workitem/workItem/state/group',
    ]
    _rr_unqueryable = [
        'workitem/workItem/type',
        'workitem/workItem/teamArea',
    ]

    def __init__(self, server, contextroot, jts=None):
        super().__init__(server, contextroot, jts=jts)
        self.rootservices_xml = self.execute_get_xml(self.reluri('rootservices'), intent="Retrieve CCM rootservices" )
        self.serviceproviders = 'oslc_cm:cmServiceProviders'

    def _get_headers(self, headers=None):
        result = super()._get_headers()
        result['net.jazz.jfs.owning-context'] = self.baseurl
        result['OSLC-Core-Version'] = '2.0'
        if headers:
            result.update(headers)
        return result

    @classmethod
    def add_represt_arguments( cls, subparsers, common_args ):
        '''
        NOTE this is called on the class (i.e. is a class method) because at this point don't know which app with be queried
        '''
        parser_ccm = subparsers.add_parser('ccm', help='CCM Reportable REST actions', parents=[common_args])
        
        parser_ccm.add_argument('artifact_format', choices=cls.artifact_formats, default=None, help=f'CCM artifact format - possible values are {", ".join(cls.artifact_formats)}')

        # SCOPE settings
        parser_ccm.add_argument('-p', '--project', default=None, help='Scope: Name of project - required when using module/collection/view/resource/typename ID/typename as a filter')

        parser_ccm.add_argument('-r', '--report', action='store_true', help='Report the fields available')

#        # Source Filters - only use one of these at once - all require a project and configuration!
#        rmex1 = parser_ccm.add_mutually_exclusive_group()
#        rmex1.add_argument('-c', '--collection', default=None, help='Sub-scope: RM: Name or ID of collection - you need to provide the project and local/global config')
#        rmex1.add_argument('-m', '--module', default=None, help='Sub-scope: RM: Name or ID of module - you need to provide the project and local/global config')
#        rmex1.add_argument('-v', '--view', default=None, help='Sub-scope: RM: Name of view - you need to provide the project and local/global config')
#        rmex1.add_argument('-r', '--resourceIDs', default=None, help='Sub-scope: RM: Comma-separated IDs of resources - you need to provide the project and local/global config')
#        rmex1.add_argument('-t', '--typename', default=None, help='Sub-scope: RM: Name of type - you need to provide the project and local/global config')
        
#        # Output FILTER settings - only use one of these at once
#        parser_ccm.add_argument('-a', '--all', action="store_true", help="Filter: Report all resources")
#        parser_ccm.add_argument('-d', '--modifiedsince', default=None, help='Filter: only return items modified since this date in format 2021-01-31T12:34:26Z')
        parser_ccm.add_argument('-f', '--fields', default=None, help="Filter using xpath")
#        parser_ccm.add_argument('-x', '--expandEmbeddedArtifacts', action="store_true", help="Filter: Expand embedded artifacts")
        
#        # various options
#    #    parser_ccm.add_argument('--forever', action='store_true', help="TESTING UNFINISHED: save web data forever (used for regression testing against stored data, may not need the target server if no requests fail)" )
#        parser_ccm.add_argument('--nresults', default=-1, type=int, help="TESTING UNFINISHED: Number of results expected (used for regression testing against stored data, doesn't need the target server - use -1 to disable checking")
#        parser_ccm.add_argument('--pagesize', default=100, type=int, help="Page size for results paging (default 100)")    
        
#        # Output controls - only use one of these at once!
        rmex2 = parser_ccm.add_mutually_exclusive_group()
#        rmex2.add_argument('--attributes', default=None, help="Output: Comma separated list of attribute names to report (requires specifying project and configuration)")
        rmex2.add_argument('--schema', action="store_true", help="Output: Report the schema")
#        rmex2.add_argument('--titles', action="store_true", help="Output: Report titles")
#        rmex2.add_argument('--linksOnly', action="store_true", help="Output: Report links only")
#        rmex2.add_argument('--history', action="store_true", help="Output: Report history")
#        rmex2.add_argument('--coverPage', action="store_true", help="Output: Report cover page variables")
#        rmex2.add_argument('--signaturePage', action="store_true", help="Output: Report signature page variables")
#    #    rmex2.add_argument('--size', action="store_true", help="Output: Set size (required for ???)")

    def process_represt_arguments( self, args, allapps ):
        '''
        Process above arguments, returning a dictionayt of parameters to add to the represt base URL
        NOTE this does have some dependency on thje overall 
        
        NOTE this is called on an instance (i.e. not a class method) because by now we know which app is being queried
        '''
        queryparams = {}
        queryurl = ""
        queryheaders={}
        
        if args.schema:
            queryparams['metadata'] = 'schema'
            
        queryurl = self.reluri(self.reportablerestbase) + "/"+ args.artifact_format
            
        if args.report:
            typestodo = []
            # get the schema, walk it building the tree of fields
            schema_x = self.execute_get_xml(queryurl+"?metadata=schema", intent="Retrieve CCM Reportable REST schema" ).getroot()
#            print( f"{schema_x.tag=}" )
#            print( f"{schema_x=}" )
            el_x = rdfxml.xml_find_element( schema_x, "./xs:element" )
            typestodo=[el_x.get('type')]
            knowntypes={el_x.get('type'):[el_x.get('type')]} # contains path to [parent
            fieldlist = []
            while typestodo:
#                print( f"{knowntypes=}" )
                typetofind = typestodo.pop(0)
                type_x = rdfxml.xml_find_element( schema_x, f'.//xs:complexType[@name="{typetofind}"]' )
#                print( f"Finding {typetofind=} {type_x=}" )
                if type_x is not None:
                    seq_x = rdfxml.xml_find_element( type_x, './xs:sequence' )
                    type_name = type_x.get('name')
                    name_name = type_x.get('type')
    #                print( f"{type_name=}" )
                    if seq_x is not None:
#                        print( f"{typetofind=} {type_name=} {name_name=}" )
                        for subel_x in rdfxml.xml_find_elements(seq_x,'./xs:element'):
                            subeltype = subel_x.get('type')
                            subelname = subel_x.get('name')
    #                        print( f"{subeltype=}" )
    #                        print( f"{subelname=}" )
                            typestr = ""
                            if subeltype.startswith( "xs:"):
#                                print( f"Type {typetofind} Found endpoint {subelname} {subeltype}" )
                                typestr = " "+subeltype                            
                                fieldpath = "/".join(knowntypes[type_name]+[subelname]) + typestr
                                if fieldpath not in fieldlist:
#                                    print( f"Adding {fieldpath}" )
                                    fieldlist.append(fieldpath)
                                
                            elif subeltype not in knowntypes:
                                typestodo.append(subeltype) 
#                                print( f"Type {typetofind} Queued {subelname=} {subeltype=}" )
                                knowntypes[subeltype] = knowntypes[type_name]+[subelname]
    #                        print( f'{"/".join(knowntypes[subeltype]+[subelname])}' )
                                
                    else:
                        raise Exception( f"xs:sequence not found in schema for type {typetofind}" )
            print( "\n".join(sorted(fieldlist) ) )

        if args.fields:
            queryparams['fields'] = args.fields

        return (queryurl,queryparams,queryheaders)



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##

import os
import logging
import datetime
import inspect
import base64
import logging
import xml.etree.ElementTree as ET

from . import rdfxml

import cryptography.fernet
import cryptography.hazmat.backends
import cryptography.hazmat.primitives
import cryptography.hazmat.primitives.kdf.pbkdf2

############################################################################
# setup logging

# from https://stackoverflow.com/a/55276759/2318649

import logging
import functools

# the TRACE level is between warning and info so the httpops can just log communication with the server without getting all the other info/debug logged
logging.TRACE = 25
logging.addLevelName(logging.TRACE, 'TRACE')
logging.Logger.trace = functools.partialmethod(logging.Logger.log, logging.TRACE)
logging.trace = functools.partial(logging.log, logging.TRACE)

#

logger = logging.getLogger(__name__)

LOGFOLDER = './logs'

loglevels = {
        'DEBUG':        logging.DEBUG
        ,'INFO':        logging.INFO
        ,'TRACE':       logging.TRACE
        ,'WARNING':     logging.WARNING
        ,'ERROR':       logging.ERROR
        ,'CRITICAL':    logging.CRITICAL
        ,'OFF':         None
        }

def setup_logging( *, filelevel=logging.INFO, consolelevel=None ):
    # make sure logs folder exists for logging output
    os.makedirs(LOGFOLDER, exist_ok=True)

    if filelevel is not None or consolelevel is not None:
        # set default logging level
        log = logging.getLogger()  # init the root logger
#        log.setLevel(logging.DEBUG)
        log.setLevel(max(filelevel or 0,consolelevel or 0))
        # make a formatter to use for the file logs
        filelogformatter = logging.Formatter("%(asctime)s [%(levelname)-5s|%(name)s] %(message)s")

        datetimestamp = '{:%Y%m%d-%H%M%S}'.format(datetime.datetime.now())

        if filelevel is not None:
            # file handler gets *all* log messages
            handler = logging.FileHandler(os.path.join(LOGFOLDER,f"elmclient-{datetimestamp}.log"), mode='w')  # create a 'log1.log' handler
            handler.setLevel(filelevel)  # make sure all levels go to it
            handler.setFormatter(filelogformatter)  # use the above formatter
            log.addHandler(handler)  # add the file handler to the root logger

        if consolelevel is not None:
            # define a Handler which writes messages to the sys.stderr
            console = logging.StreamHandler()
            console.setLevel(consolelevel)
            # set a format which is simpler for console use
            formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
            # tell the handler to use this format
            console.setFormatter(formatter)
            # add the handler to the root logger
            log.addHandler(console)
            
############################################################################

def log_commandline( prog,args=None ):
    args = args or []
    def optquote(s):
        if " " in s:
            return '"s"'
        return s
    arg = " ".join([optquote(a) for a in args])
    logger.trace( f"COMMANDLINE: {prog} {arg}" )

############################################################################

def log_state( statename ):
    logger.trace( f"STATE: {statename}" )

############################################################################
# return string for the n
def nth( n ):
    if n < 1:
        raise Exception( f"nth can't figure out what to return for {n} - should be greater than 0!" )
    if n==1:
        return "1st"
    elif n==2:
        return "2nd"
    elif n==3:
        return "3rd"
    return f"{str(n)}th"
    
############################################################################
# code to support obfuscated credentials files

ITERATIONS = 100000
TTL = 28*24*60*60

def _derive_key( password, salt, iterations=ITERATIONS ):
    kdf =  cryptography.hazmat.primitives.kdf.pbkdf2.PBKDF2HMAC(
        algorithm=cryptography.hazmat.primitives.hashes.SHA256(), length=32, salt=salt,
        iterations=iterations, backend=cryptography.hazmat.backends.default_backend())
    return base64.urlsafe_b64encode(kdf.derive(password))

def fernet_encrypt( message, password, iterations=ITERATIONS ) :
    salt = os.urandom(16)
    key = _derive_key(password.encode(), salt, iterations)
    return base64.urlsafe_b64encode( b'%b%b%b' % ( salt, iterations.to_bytes(4, 'big'), base64.urlsafe_b64decode( cryptography.fernet.Fernet(key).encrypt(message)), ) )

def fernet_decrypt( token, password, ttl=TTL ):
    decoded = base64.urlsafe_b64decode(token)
    salt, iter, token = decoded[:16], decoded[16:20], base64.urlsafe_b64encode(decoded[20:])
    iterations = int.from_bytes(iter, 'big')
    key = _derive_key(password.encode(), salt, iterations)
    return  cryptography.fernet.Fernet(key).decrypt(token,ttl=ttl)

############################################################################
#
# For Reportable REST
#
# visit an xml tag hierarchy, extracting values into rows, one per first-level tag
# this is aimed at Reportable REST XML output
#
# to avoid duplicate names and map clearly to the XML:
#   the path to a tag is turned into /-separated tag path
#   at a level, all attributes of a tag are added to the tag path with "-attributename"
#
def getcontentrow( node, remove_ns=True ):
    thisrowdict = {}
    allcolumns=[]
    path=""
    row,columns = getacontentrow( node, thisrowdict, allcolumns, 1, path, remove_ns=remove_ns)
    return row

def getacontentrow( node, thisrowdict, allcolumns, level, path, remove_ns=True, merge_duplicates=False ):
    logger.debug( f"2 {node=} {allcolumns=} {thisrowdict=}" )
    children = list(node)

    # ensure path is unique
    if path in allcolumns:
        # add a digit to get a unique path
        for i in range(1000):
            if f"{path}{i}" not in allcolumns:
                break
        path = f"{path}{i}"
        logger.debug( f"unique1 ============================= {path=}" )
        
    if path not in allcolumns:
        allcolumns.append(path)
        
    logger.debug( f"{path=}" )
    # record attributes of node
    for k in node.keys():
        # work out the path to this attribute
        if remove_ns and '}' in k:
            k1 = k.split('}',1)[1]
        else:
            k1 = k

        if path:
            kpath = f"{path}-{k1}"
        else:
            kpath = k1
            
        # ensure the path is remembered in allcolumns
        if kpath not in allcolumns:
            allcolumns.append(kpath)
        else:
            logger.debug( f"{kpath=}" )
            raise Exception( "Unexpected {kpath} not in allcolumns" )
            
        # store the value into the row
        thisrowdict[kpath] = node.get(k,"")

    text = (node.text or "").strip()
    tail = (node.tail or "").strip()

    if tail:
        raise Exception("XML has tail - can't handle this!")

    if text or node.tag==f'{{{rdfxml.RDF_DEFAULT_PREFIX["rm_text"]}}}richTextBody':            
        if len(children)>0 and node.tag==f'{{{rdfxml.RDF_DEFAULT_PREFIX["rm_text"]}}}richTextBody':
            # this is a special case - this tag with children contains literal XHTML which we want
            # to use as-is so copy the string version of the text content and don't recurse into it
            thisrowdict[path] = ET.tostring(node)
        else:
            # this is just all the the content
            thisrowdict[path]=text+tail
        # remember paths that have a value stored
    elif len(children)>0:
        # recurse into the children
        for child in children:
            if remove_ns and '}' in child.tag:
                thistag = child.tag.split('}',1)[1]
            else:
                thistag=child.tag
            subpath = path + "/" + thistag if path !='' else thistag
            # recurse
            getacontentrow( child, thisrowdict, allcolumns, level + 1, subpath, remove_ns=remove_ns)
    else:
        # empty tag
        pass

    return (thisrowdict,allcolumns)

#####################################################################################
# return function stack trace of the calling line, as a string of file:function:line<=file:function:ine<=...
def callers():
    caller_list = []
    frame = inspect.currentframe().f_back.f_back
    while frame.f_back:
        caller_list.append('{2}:{1}:{0}()'.format(frame.f_code.co_name,frame.f_lineno,frame.f_code.co_filename.split("\\")[-1]))
        frame = frame.f_back
    callers =  ' <= '.join(caller_list)
    return callers

#####################################################################################
# based on https://stackoverflow.com/a/12065663/2318649
def print_in_columns( rows ):
    result = ""
    colcount = max([len(row) for row in rows])
    rows = [row + ['']*(colcount-len(row)) for row in rows]
    widths = [max(map(len, col)) for col in zip(*rows)]
    # ensure the final column doesn't have spaces added on the end
    widths[-1]=0
    for row in rows:
        result += "  ".join((val.ljust(width) for val, width in zip(row, widths)))+"\n"
    return result

#####################################################################################
# based on https://stackoverflow.com/a/12065663/2318649
def print_in_html( rows,headings=None ):
    result = ""
    if headings is None:
        headings = []
    colcount = max([len(row) for row in rows])
    # extend all the rows
    rows = [row + ['']*(colcount-len(row)) for row in rows]
    # extend the headings
    headings += ['']*(colcount-len(headings))
    result += "<TABLE><THEADING>\n"
    for heading in headings:
        result += f"<TH>{heading}</TH>"
    result += "</THEADING>\n"
    for row in rows:
        result += "  <TR>"
        for col in row:
            result += f"    <TD>{col}</TD>\n"
        result += "  </TR>\n"
    result += "</TABLE>\n"
    return result

#####################################################################################
# decorator for users of mixin classes - ensures their __init__ gets called
# based on https://stackoverflow.com/a/6100595/2318649
def mixinomatic(cls):
    """ Mixed-in class decorator. """
    classinit = cls.__dict__.get('__init__')  # Possibly None.

    # Define an __init__ function for the class.
    def __init__(self, *args, **kwargs):
        # Call the __init__ functions of all the bases.
        for base in cls.__bases__:
            if base is object or ( len(args)==0 and len(kwargs)==0): # added, doesn't pass args to object
                base.__init__(self)
            else:
                logger.debug( f"{base=} {args=} {kwargs=}" )
                base.__init__(self, *args, **kwargs)
        # Also call any __init__ function that was in the class.
        if classinit:
            classinit(self, *args, **kwargs)

    # Make the local function the class's __init__.
    setattr(cls, '__init__', __init__)
    return cls

############################################################################

def isint(s ):
    try:
        i = int(s)
    except:
        return None
    return i


##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


# example of using the elmclient package to download a work item attachment

import logging

import lxml.etree as ET

import elmclient.server as elmserver
import elmclient.utils as utils
import elmclient.rdfxml as rdfxml

logger = logging.getLogger(__name__)

jazzhost = 'https://jazz.ibm.com:9443'
username = 'ibm'
password = 'ibm'

jtscontext = 'jts' # specifies /jts change if your jts is on a different context root
ccmcontext  = 'ccm' # specifies /ccm change if your ccm is on a different context root

workitemid=38

# caching control
# 0=fully cached (but code below specifies queries aren't cached)
# 1=clear cache initially then continue with cache enabled
# 2=clear cache and disable caching
caching = 0

# create our "server" which is how we connect to DOORS Next
# first enable the proxy so if a proxy is running it can monitor the communication with server (this is ignored if proxy isn't running)
elmserver.setupproxy(jazzhost,proxyport=8888)
theserver = elmserver.JazzTeamServer(jazzhost, username, password, verifysslcerts=False, jtsappstring=f"jts:{jtscontext}", appstring='rm', cachingcontrol=caching)

# create the CMM application interface
ccmapp = theserver.find_app( f"ccm:{ccmcontext}", ok_to_create=True )

wiurl = ccmapp.reluri(f'resource/itemName/com.ibm.team.workitem.WorkItem/{workitemid}')

print( f"Retrieving wi details" )

xmlresult = ccmapp.server.execute_get_xml(wiurl, cacheable=False, headers={'OSLC-Core-Version': '2.0'}, intent="Retrieve work item content (including attachment details)" )

print( f"{xmlresult=}" )

for attachment in rdfxml.xml_find_elements( xmlresult, './/rtc_cm:com.ibm.team.workitem.linktype.attachment.attachment' ):
    # find the attachment URI
    # <rtc_cm:com.ibm.team.workitem.linktype.attachment.attachment oslc_cm:collref="https://jazz.ibm.com:9443/ccm/oslc/workitems/_SPJOV0B2EeuDvqpPB-fV1g/rtc_cm:com.ibm.team.workitem.linktype.attachment.attachment"/>
    attachment_u = rdfxml.xmlrdf_get_resource_uri( attachment )
    print( f"{attachment_u=}" )

    # get the attachment metadata
    attachment_info = ccmapp.server.execute_get_xml(attachment_u, cacheable=False, headers={'OSLC-Core-Version': '2.0'}, intent="Retrieve attachment details (points to attachment content)" )

    # find the download link
    # <rtc_cm:content rdf:resource="https://jazz.ibm.com:9443/ccm/resource/content/_IZCsIRuREeyjc_YwJfTLJA"/>
    download_u = rdfxml.xmlrdf_get_resource_uri( attachment_info, './/rtc_cm:content' )
    print( f"{download_u=}" )
    filename = rdfxml.xmlrdf_get_resource_text( attachment_info, './/dcterms:title' )
    print( f"{filename=}" )
    # download it, using Referer set to the URI (addresses a security measure built-in to ccm)
    attachment_binary = ccmapp.server.execute_get_binary(download_u, cacheable=False, headers={'OSLC-Core-Version': '2.0', 'Referer':download_u}, intent="Retrieve attachement content (binary)" )
    print( f"{len(attachment_binary.content)}" )
    # open( filename,"wb" ).write( attachment_binary )

print( "Finished" )



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##

# example of using the elmclient package as a DOORS Next API to do reportable rest query
# exporting to CSV

# NOTE accessing the Reportable REST API may add a lot of load to your server so use with care!
# because it tries to retrieve *all* DOORS Next artifacts!
# code below only retrieves the first two pages!

# DN/DNG Reportable REST API https://jazz.net/wiki/bin/view/Main/DNGReportableRestAPI
#   DN/DNG FAQ on reportable rest https://jazz.net/wiki/bin/view/Main/RRCReportServiceFAQ
# ETM/RQM Reportable REST API https://jazz.net/wiki/bin/view/Main/RqmApi#Resources_and_their_Supported_Op
# EWM/RTC Reportable REST API https://jazz.net/wiki/bin/view/Main/ReportsRESTAPI

import csv
import logging

import lxml.etree as ET

import elmclient.server as elmserver
import elmclient.utils as utils

logger = logging.getLogger(__name__)

jazzhost = 'https://jazz.ibm.com:9443'
username = 'ibm'
password = 'ibm'

jtscontext = 'jts'
rmcontext  = 'rm'

outfile = "dnrr.csv"

# caching control
# 0=fully cached (but code below specifies queries aren't cached)
# 1=clear cache initially then continue with cache enabled
# 2=clear cache and disable caching
caching = 0

# create our "server" which is how we connect to DOORS Next
# first enable the proxy so if a proxy is running it can monitor the communication with server (this is ignored if proxy isn't running)
elmserver.setupproxy(jazzhost,proxyport=8888)
theserver = elmserver.JazzTeamServer(jazzhost, username, password, verifysslcerts=False, jtsappstring=f"jts:{jtscontext}", appstring='rm', cachingcontrol=caching)

# create the RM application interface
dnapp = theserver.find_app( f"rm:{rmcontext}", ok_to_create=True )

# A DN Reportable Rest query url (this also works in your browser to get the first page of results)
rrurl = dnapp.reluri( dnapp.reportablerest_baseurl+"/resources/*" )
print( f"{rrurl=}" )

# limit to this number of pages to limit load on server
# the Reportable REST query "resources/*" used here, with no constraining project or module URI will attempt
# to retrieve all the artifacts in the default configuration of all projects
# that can be A LOT of data even on a very small deployment.
pagelimit = 3

# retrieve all the results - works for one or many pages
rows = []
allcolumns = []

print( f"Retrieving {pagelimit} pages from DOORS Next Reportable REST API" )

headings = []
rows=[]

while rrurl is not None and pagelimit > 0:
    # perform the Reportable Rest query
    # because query results are often updated as users work, this request is NOT cached even if caching is enabled
    print( f"Requesting page {rrurl}" )
    xmlresult = dnapp.execute_get_xml(rrurl, cacheable=False, remove_headers=['net.jazz.jfs.owning-context'], intent="Retrieve a page of Reportable REST results" )
    root = xmlresult.getroot()

    # go to each result (one row) below the root  -this will be a row of the results
    for res in list(xmlresult.getroot()):
        row = {}
        # def getcontentrow( node, thisrowdict, allcolumns, level, path, remove_ns=True ):
        row = utils.getcontentrow( res )
        for k in row.keys():
            if k not in headings:
                headings.append(k)                
        rows.append(row)
        
    # find the href for the next page of results
    rrurl = root.get("href",None)
    pagelimit -= 1

print( f"Writing to CSV {outfile}" )

with open( outfile, "w", newline='' ) as csvfile:
    csvwriter = csv.DictWriter(csvfile,fieldnames=sorted(headings))
    csvwriter.writeheader()
    for row in rows:
        csvwriter.writerow(row)

print( "Finished" )



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


#################################################################################################

# GCM API documentation https://jazz.net/gc/doc/scenarios

#
# This is a skeleton implementation for GC - just enough to support finding a GC configuration (using OSLC query to GC) either in a project or across all projects
#

#################################################################################################

import logging

import requests
import lxml.etree as ET
import tqdm

from . import _app
from . import _project
from . import _typesystem
from . import oslcqueryapi
from . import rdfxml
from . import server
from . import utils

logger = logging.getLogger(__name__)

#################################################################################################

# hook to adapt OSLC query parameters needed for GC - no orderBy, prefixes must NOT include dcterms
def _hook_beforequery(querydetails):
    # remove orderby
    if 'orderBy' in querydetails:
        del querydetails['orderBy']
    # make sure dcterms and oslc not in prefix
    if 'dcterms=' in querydetails.get('oslc.prefix',"") or 'oslc' in querydetails.get('oslc.prefix',"") or 'rdf' in querydetails.get('oslc.prefix',""):
        oldprefix = querydetails['oslc.prefix']
        prefixes = oldprefix.split(",")
        newprefixes = [p for p in prefixes if not p.startswith("dcterms=") and not p.startswith("oslc=") and not p.startswith("rdf=")]
        querydetails['oslc.prefix'] = ",".join(newprefixes)
        newprefix = querydetails['oslc.prefix']
    return querydetails

#################################################################################################

class _GCMProject(_project._Project):
    def __init__(self, name, project_uri, app, is_optin=False, singlemode=False,defaultinit=True):
        super().__init__(name, project_uri, app, is_optin,singlemode, defaultinit=False)
        self.hooks = [_hook_beforequery]
        self._components = None  # will become a dict keyed on component uri
        self._configurations = None # keyed on the config name
        self.default_query_resource = 'oslc_config:Configuration'

        # nonstandard initialisation - difference is in finding the services XML for the project - for
        # some reason GCM uses oslc:serviceProvider and doesn't include the project name whereas at
        # least RM uses oslc:ServiceProvider! and does include the project name - that's what the default initialisation assumes

        # we could retrieve each of these to check the title, but using the project iid to find the correct services xml URL works well enough

        # get the app oslc catalog (this is different from the other apps!)
        appcatalog_xml = self.app.retrieve_oslc_catalog_xml()
        self.services_uri = None
        for sp in rdfxml.xml_find_elements(appcatalog_xml, ".//oslc:serviceProvider"):
            spuri = rdfxml.xmlrdf_get_resource_uri( sp )
            if spuri.endswith( self.iid):
                self.services_uri = spuri
                break

        if not self.services_uri:
            raise Exception( "Service provide not found!" )
        if self.services_uri:
            self.services_xml = self.app.execute_get_rdf_xml( self.services_uri, intent="Retrieve project services xml" )
        else:
            self.services_xml = None

    def find_local_component(self, name_or_uri):
        self.load_components_and_configurations()
        for compuri, compdetail in self._components.items():
            logger.info( f"Checking {name_or_uri} {compdetail}" )
            if compuri == name_or_uri or compdetail['name'] == name_or_uri:
                return compdetail['component']
        return None

    def load_components_and_configurations(self,force=False):
        if self._components is not None and self._configurations is not None and not force:
            return
        self._components = {}
        self._configurations = {}
        ncomps = 0
        nconfs = 0
        # retrieve components and configurations for this project
        # use OSLC query to get all components in this project
        comps = self.do_complex_query( "http://open-services.net/ns/config#Component", querystring=None, select='dcterms:title', show_progress=False, verbose=False )

        logger.debug( f"{comps=}" )

        for compu,v in comps.items():
            self._components[compu] = {'name': v['dcterms:title'], 'configurations': {}}
            logger.debug( f"{self._components[compu]=}" )

        # now create the "components"
        for cu, cd in self._components.items():
            logger.debug( f"{cu=} {cd=}" )
            cname = cd['name']
            c = self
            c._configurations = self._components[cu]['configurations']
            self._components[cu]['component'] = c
        return (ncomps, nconfs)

    # load the typesystem using the OSLC shape resources
    def _load_types(self,force=False):
        logger.debug( f"load type {self=} {force=}" )

        # if already loaded, try to avoid reloading
        if self.typesystem_loaded and not force:
            return

        self.clear_typesystem()

        if self.local_config:
            # get the configuration-specific services.xml
            sx = self.get_services_xml(force=True,headers={'configuration.Context': self.local_config, 'net.jazz.jfs.owning-context': None})
        else:
            # No config - get the services.xml
            sx = self.get_services_xml(force=True)
        if sx:
            shapes_to_load = rdfxml.xml_find_elements(sx, './/oslc:resourceShape')

            pbar = tqdm.tqdm(initial=0, total=len(shapes_to_load),smoothing=1,unit=" results",desc="Loading GCM project shapes")

            for el in shapes_to_load:
                self._load_type_from_resource_shape(el)
                pbar.update(1)

            pbar.close()
        else:
            raise Exception( "services xml not found!" )

        self.typesystem_loaded = True
        burp
        return None

# pick all the attributes from a resource shape definition
    # and for enumerated attributes get all the enumeration values
    def _load_type_from_resource_shape(self, el, supershape=None):
        return self._generic_load_type_from_resource_shape(el, supershape=None)

    def _generic_load_type_from_resource_shape(self, el, supershape=None):
        logger.debug( "Starting a shape")
        uri = rdfxml.xmlrdf_get_resource_uri(el)
        try:
            if not self.is_known_shape_uri(uri):
                logger.info( f"Starting shape {uri} =======================================" )
                logger.debug( f"Getting {uri}" )
                shapedef = self._get_typeuri_rdf(uri)
                # find the title
                name_el = rdfxml.xml_find_element(shapedef, f'.//rdf:Description[@rdf:about="{uri}"]/dcterms:title[@xml:lang="en"]')
                if name_el is None:
                    name_el = rdfxml.xml_find_element(shapedef, f'.//rdf:Description[@rdf:about="{uri}"]/dcterms:title')
                if name_el is None:
                    name = uri.rsplit('#',1)[1]
                    logger.info( f"MADE UP NAME {name}" )
                else:
#                    print( "NO NAME",ET.tostring(shapedef) )
#                    raise Exception( "No name element!" )
                    name = name_el.text
                self.register_shape( name, uri )
                logger.info( f"Opening shape {name} {uri}" )
            else:
                return
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                logger.info( f"Failed because type not found 404 - ignoring! {e}")
                return
            elif e.response.status_code == 410:
                logger.info( f"Failed because type not found 410 - ignoring! {e}")
                return
            else:
                raise

        n = 0
        # find the list of attributes
        thisshapedef = rdfxml.xml_find_element( shapedef,f'.//rdf:Description[@rdf:about="{uri}"]' )
        if thisshapedef is None:
            raise Exception( f"Shape definition for {uri} not found!" )
#        print( f"thisshapedef=",ET.tostring(thisshapedef) )
        title = rdfxml.xmlrdf_get_resource_text(thisshapedef,'./dcterms:title[@xml:lang="en"]')
        if title is None:
            title = rdfxml.xmlrdf_get_resource_text(thisshapedef,'./dcterms:title')

#        logger.info( f"shape {title} xml={ET.tostring(thisshapedef)}" )
        # scan the attributes
        for propel in rdfxml.xml_find_elements( thisshapedef,'./oslc:property' ):
            logger.info( "Starting a property")
            propnodeid = rdfxml.xmlrdf_get_resource_uri(propel,attrib="rdf:nodeID")
            logger.info( f"{propnodeid=}" )
            real_propel = rdfxml.xml_find_element(shapedef, f'.//rdf:Description[@rdf:nodeID="{propnodeid}"]')
            logger.info( f"{real_propel=}" )
#            print( "XML==",ET.tostring(real_propel) )
            # dcterms:title xml:lang="en"
            property_title_el = rdfxml.xml_find_element(real_propel, './dcterms:title[@xml:lang="en"]')
            if property_title_el is None:
                property_title_el = rdfxml.xml_find_element(real_propel, './dcterms:title')
            logger.info( f"{property_title_el=}" )
            if property_title_el is None:
                logger.info( "Skipping shape with no title!" )
                continue
            property_title = property_title_el.text
            logger.info( f"{property_title=}" )
            if rdfxml.xmlrdf_get_resource_text(propel,"oslc:hidden") == "true":
                logger.info( f"Skipping hidden property {property_title}" )
                continue
            valueshape_uri = rdfxml.xmlrdf_get_resource_uri(real_propel,'./oslc:valueShape')
            if valueshape_uri is not None:
                logger.info( f"vs {valueshape_uri}" )
                # register this property with a fake URI using the node id
                propu = f"{uri}#{propnodeid}"
                self.register_property( property_title, propu, shape_uri=uri )
                if valueshape_uri.startswith( self.app.baseurl):
                    # this shape references another shape - need to load this!
                    vs_xml = self._get_typeuri_rdf(valueshape_uri)
                    subshape_x = rdfxml.xml_find_element( vs_xml,f'.//rdf:Description[@rdf:about="{valueshape_uri}"]' )
                    if subshape_x is None:
#                        print( f"\n\nSUBSHAPE_X=",ET.tostring( vs_xml),"\n\n" )
                        logger.info( f"SubShape definition for {valueshape_uri} not found!" )
                        continue
                    # recurse to load this shape!
                    self._load_type_from_resource_shape( subshape_x, supershape=(property_title,propu))
                else:
                    logger.info( f"SKIPPED external shape {valueshape_uri=}" )
            else:
                logger.debug( f"{valueshape_uri=}" )
#                if valueshape_uri is None or not valueshape_uri.startswith( self.baseurl):
#                    if not valueshape_uri:
#                        valueshape_uri = f"{uri}#{propnodeid}"
#                    if not self.is_known_property_uri( valueshape_uri ):
#                        # define it
#                        self.register_property( property_title, valueshape_uri, shape_uri=uri )
#                        continue
#                    else:
#                        logger.debug( f"ALREADY KNOWN" )
#                        continue

                pd_u = rdfxml.xmlrdf_get_resource_uri( real_propel, 'oslc:propertyDefinition' )

                # In case of repeated identical property titles on a shape, let's create an alternative name that can (perhaps) be used to disambiguate
                # (at least these don't have duplicates AFAICT)
                altname  = pd_u[pd_u.rfind("/")+1:]
                if '#' in altname:
                    altname  = pd_u[pd_u.rfind("#")+1:]

                if pd_u is not None:
                    if not pd_u.startswith( self.app.baseurl ):
                        self.register_property(property_title,pd_u, shape_uri=uri, altname=altname)
                        logger.debug( f"+++++++Skipping non-local Property Definition {pd_u}" )
                        continue
                else:
                    logger.debug( f"~~~~~~~Ignoring non-local Property Definition {pd_u}" )

                if self.is_known_property_uri( pd_u,shape_uri=uri,raiseifnotfound=False ):
                    logger.debug( f"ALREADY KNOWN2" )
                    continue

                logger.info( f"Defining property {title}.{property_title} {altname=} {pd_u=} +++++++++++++++++++++++++++++++++++++++" )
                self.register_property(property_title,pd_u, shape_uri=uri, altname=altname)
                # check for any allowed value
                allowedvalueu = rdfxml.xmlrdf_get_resource_uri(real_propel, ".//oslc:allowedValue" )
                if allowedvalueu is not None:
                    logger.info( "FOUND ENUM" )
                    # this has enumerations - find them and record them
                    # retrieve each definition
                    nvals = 0
                    for allowedvaluex in rdfxml.xml_find_elements( real_propel,'.//oslc:allowedValue'):
                        allowedvalueu = rdfxml.xmlrdf_get_resource_uri(allowedvaluex )

                        thisenumx = rdfxml.xml_find_element( shapedef,f'.//rdf:Description[@rdf:about="{allowedvalueu}"]' )

                        enum_uri = allowedvalueu
                        logger.info( f"{enum_uri=}" )
                        nvals += 1
                        if not self.is_known_enum_uri( enum_uri ):
                            # retrieve it and save the enumeration name and uri in types cache
                            enum_value_name = rdfxml.xmlrdf_get_resource_text(thisenumx, 'rdfs:label')
                            enum_id = enum_value_name
                            if enum_value_name is None:
                                logger.debug( "enum xml=",ET.tostring(thisenumx) )
                                logger.debug( f"{enum_id=} no name" )
                                raise Exception( "Enum name not present!" )

                            logger.info( f"defining enum value {enum_value_name=} {enum_id=} {enum_uri=}" )
                            self.register_enum( enum_value_name, enum_uri, property_uri=pd_u, id=None )

                    if nvals==0:
                        raise Exception( f"Enumeration {valueshape_uri} with no values loaded" )
        logger.debug( "Finished loading typesystem")
        return n

    # this is a local cache only for the typesystem retrieval
    # the cache deliberately strips off the fragment because it's irrelevant for the GET
    # this makes a lot of the URIs repeats so they are cached
    def _get_typeuri_rdf(self,uri):
        # strip off the fragment as it does nothing
        realuri = uri.rsplit( '#',1 )[0]
        if realuri not in self._gettypecache.keys():
            self._gettypecache[realuri] = self.execute_get_rdf_xml( uri, intent="Retrieve project/component type definition to cache it" )
        return self._gettypecache[realuri]

    # for OSLC query, given a type URI, return its name
    # gcm-specific resolution
    def app_resolve_uri_to_name(self, uri):
        result = None
        return result

    # given a type URI, return its name
    def resolve_uri_to_name(self, uri, prefer_same_as=True, dontpreferhttprdfrui=True):
        logger.info( f"resolve_uri_to_name {uri=}" )
        if not uri:
            result = None
            return result
        if not uri.startswith('http://') or not uri.startswith('https://'):
        # try to remove prefix
            uri1 = rdfxml.tag_to_uri(uri,noexception=True)
            logger.debug(f"Trying to remove prefix {uri=} {uri1=}")
            if uri1 is None:
                return uri
            if uri1 != uri:
                logger.debug( f"Changed {uri} to {uri1}" )
            else:
                logger.debug( f"NOT Changed {uri} to {uri1}" )
            # use the transformed URI
            uri = uri1
        if not uri.startswith(self.reluri()):
            if self.server.jts.is_user_uri(uri):
                result = self.server.jts.user_uritoname_resolver(uri)
                logger.debug(f"returning user")
                return result
            uri1 = rdfxml.uri_to_prefixed_tag(uri,noexception=True)
            logger.debug(f"No app base URL {self.reluri()=} {uri=} {uri1=}")
            return uri1
        elif not self.is_known_uri(uri):
            if self.server.jts.is_user_uri(uri):
                result = self.server.jts.user_uritoname_resolver(uri)
            else:
                if uri.startswith( "http://" ) or uri.startswith( "https://" ):
                    uri1 = rdfxml.uri_to_prefixed_tag(uri)
                    logger.debug( f"Returning the raw URI {uri} so changed it to prefixed {uri1}" )
                    uri = uri1
                result = uri
            # ensure the result is in the types cache, in case it recurrs the result can be pulled from the cache
            self.register_name(result,uri)
        else:
            result = self.get_uri_name(uri)
        logger.info( f"Result {result=}" )
        return result


#################################################################################################

class _GCMComponent(_GCMProject):
    pass

#################################################################################################

@utils.mixinomatic
class GCMApp(_app._App, oslcqueryapi._OSLCOperations_Mixin, _typesystem.Type_System_Mixin):
    domain = 'gc'
    project_class = _GCMProject
    supports_configs = False
    supports_components = True
    supports_reportable_rest = False

    relprefixes = (
            ('acclist','acclist#')
            ,('serviceProvider','oslc-config/serviceProvider/')
            ,('component','component/')
            ,('configuration','configuration/')
            ,('part','part/')
        )
    identifier_name = 'Short ID'
    identifier_uri = 'Identifier'

    def __init__(self, server, contextroot, jts=None):
        super().__init__(server, contextroot, jts=jts)

        self.rootservices_xml = self.execute_get_xml(self.reluri('rootservices'), intent="Retrieve GCM application rootservices" )
        self.serviceproviders = 'gc:globalConfigServiceProviders'
        self.default_query_resource = 'oslc_config:Configuration'
        # load all projects and components?
        # register some app-specific namespaces
        for prefix,reluri in self.relprefixes:
            rdfxml.addprefix(prefix,self.baseurl+reluri)
        self.hooks = [_hook_beforequery]
        self.iid = None
#        self._type_system = _typesystem.Type_System()
#        self._gettypecache = {}

    def _get_headers(self, headers=None):
        result = super()._get_headers()
        result['net.jazz.jfs.owning-context'] = self.baseurl
        if headers:
            result.update(headers)
        return result

    # returns a dictionary of resource type to query capability URI
    # this is here because for the GCM app the xml is structured differently for the app
    def get_query_capability_uris_from_xml(self,capabilitiesxml,context):
        logger.info( f"GCM gqcusfx" )
        qcs = {}
        if context.iid:
            # project
            #<oslc:QueryCapability>
            #    <oslc:resourceType rdf:resource="http://open-services.net/ns/cm#ChangeRequest"/>
            #    <oslc:queryBase rdf:resource="https://jazz.ibm.com:9443/ccm/oslc/contexts/_2H-_4OpoEemSicvc8AFfxQ/workitems"/>
            # find a queryBase and it's the containing tag that has the info
            for qcx in rdfxml.xml_find_elements(capabilitiesxml,'.//oslc:queryBase/..'):
                for qcrtx in rdfxml.xml_find_elements( qcx, 'oslc:resourceType'):
                    qcs[rdfxml.xmlrdf_get_resource_uri(qcrtx)] = rdfxml.xmlrdf_get_resource_uri(qcx, "oslc:queryBase")
                    logger.debug( f"{rdfxml.xmlrdf_get_resource_uri(qcrtx)=}" )
                    logger.info( f"GCM Project get_query_capability_uris_from_xml returning {qcs=}" )
        else:
            # app
            sps = rdfxml.xml_find_elements(capabilitiesxml,".//oslc:serviceProvider")
            found=False
            for sp in sps:
                spurl = rdfxml.xmlrdf_get_resource_uri(sp)
                if spurl.endswith("serviceProvider"):
                    found=True
                    break
            if not found:
                raise Exception( "No empty service provider found!" )
            sx = self.execute_get_rdf_xml( spurl, intent="Retrieve project/component service provider definition to find query capability" )
            for qcx in rdfxml.xml_find_elements(sx,'.//oslc:queryBase/..'):
                for qcrtx in rdfxml.xml_find_elements( qcx, 'oslc:resourceType'):
                    qcs[rdfxml.xmlrdf_get_resource_uri(qcrtx)] = rdfxml.xmlrdf_get_resource_uri(qcx, "oslc:queryBase")
                    logger.debug( f"{rdfxml.xmlrdf_get_resource_uri(qcrtx)=}" )
            logger.info( f"GCM App get_query_capability_uris_from_xml returning {qcs=}" )

        return qcs

    def check_valid_config_uri( self, uri, raise_exception=True ):
        try:
            x = self.rootservices_xml = self.execute_get_xml( uri, intent="Check if configuration URL is valid (gets a response)" )
        except requests.HTTPError:
            if raiseException:
                raise
            return False
        return True


    # load the typesystem using the OSLC shape resources listed for all the creation factories and query capabilities
    def load_types(self, force=False):
        self._load_types(force)

    # load the typesystem using the OSLC shape resources listed for all the creation factories and query capabilities
    def _load_types(self,force=False):
        logger.debug( f"load type {self=} {force=}" )
        # if already loaded, try to avoid reloading
        if self.typesystem_loaded and not force:
            return

        self.clear_typesystem()

        # app
        capabilitiesxml = self.retrieve_cm_service_provider_xml()
        sps = rdfxml.xml_find_elements(capabilitiesxml,".//oslc:serviceProvider")
        found=False
        for sp in sps:
            spurl = rdfxml.xmlrdf_get_resource_uri(sp)
            if spurl.endswith("serviceProvider"):
                found=True
                break
        if not found:
            raise Exception( "No empty service provider found!" )
        sx = self.execute_get_rdf_xml( spurl, intent="Retrieve project/component service provider XML" )
        if sx:
            shapes_to_load = rdfxml.xml_find_elements(sx, './/oslc:resourceShape' )
            pbar = tqdm.tqdm(initial=0, total=len(shapes_to_load),smoothing=1,unit=" results",desc="Loading GCM app shapes")

            for el in shapes_to_load:
                self._load_type_from_resource_shape(el)
                pbar.update(1)

            pbar.close()

        else:
            raise Exception( "services xml not found!" )
        return None

    # pick all the attributes from a resource shape definition
    # and for enumerated attributes get all the enumeration values
    def _load_type_from_resource_shape(self, el, supershape=None):
        return self._generic_load_type_from_resource_shape(el, supershape=None)

    def _generic_load_type_from_resource_shape(self, el, supershape=None):
        logger.debug( "Starting a shape")
        uri = rdfxml.xmlrdf_get_resource_uri(el)
        try:
            if not self.is_known_shape_uri(uri):
                logger.info( f"Starting shape {uri} =======================================" )
                logger.debug( f"Getting {uri}" )
                shapedef = self._get_typeuri_rdf(uri)
                # find the title
                name_el = rdfxml.xml_find_element(shapedef, f'.//rdf:Description[@rdf:about="{uri}"]/dcterms:title[@xml:lang="en"]')
                if name_el is None:
                    name_el = rdfxml.xml_find_element(shapedef, f'.//rdf:Description[@rdf:about="{uri}"]/dcterms:title')
                if name_el is None:
                    name = uri.rsplit('#',1)[1]
                    logger.info( f"MADE UP NAME {name}" )
                else:
#                    print( "NO NAME",ET.tostring(shapedef) )
#                    raise Exception( "No name element!" )
                    name = name_el.text
                self.register_shape( name, uri )
                logger.info( f"Opening shape {name} {uri}" )
            else:
                return
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                logger.info( f"Failed because type not found 404 - ignoring! {e}")
                return
            elif e.response.status_code == 410:
                logger.info( f"Failed because type not found 410 - ignoring! {e}")
                return
            else:
                raise

        n = 0
        # find the list of attributes
        thisshapedef = rdfxml.xml_find_element( shapedef,f'.//rdf:Description[@rdf:about="{uri}"]' )
        if thisshapedef is None:
            raise Exception( f"Shape definition for {uri} not found!" )
#        print( f"thisshapedef=",ET.tostring(thisshapedef) )
        title = rdfxml.xmlrdf_get_resource_text(thisshapedef,'./dcterms:title[@xml:lang="en"]')
        if title is None:
            title = rdfxml.xmlrdf_get_resource_text(thisshapedef,'./dcterms:title')

#        logger.info( f"shape {title} xml={ET.tostring(thisshapedef)}" )
        # scan the attributes
        for propel in rdfxml.xml_find_elements( thisshapedef,'./oslc:property' ):
            logger.info( "Starting a property")
            propnodeid = rdfxml.xmlrdf_get_resource_uri(propel,attrib="rdf:nodeID")
            logger.info( f"{propnodeid=}" )
            real_propel = rdfxml.xml_find_element(shapedef, f'.//rdf:Description[@rdf:nodeID="{propnodeid}"]')
            logger.info( f"{real_propel=}" )
#            print( "XML==",ET.tostring(real_propel) )
            # dcterms:title xml:lang="en"
            property_title_el = rdfxml.xml_find_element(real_propel, './dcterms:title[@xml:lang="en"]')
            if property_title_el is None:
                property_title_el = rdfxml.xml_find_element(real_propel, './dcterms:title')
            logger.info( f"{property_title_el=}" )
            if property_title_el is None:
                logger.info( "Skipping shape with no title!" )
                continue
            property_title = property_title_el.text
            logger.info( f"{property_title=}" )
            if rdfxml.xmlrdf_get_resource_text(propel,"oslc:hidden") == "true":
                logger.info( f"Skipping hidden property {property_title}" )
                continue
            valueshape_uri = rdfxml.xmlrdf_get_resource_uri(real_propel,'./oslc:valueShape')
            if valueshape_uri is not None:
                logger.info( f"vs {valueshape_uri}" )
                # register this property with a fake URI using the node id
                propu = f"{uri}#{propnodeid}"
                self.register_property( property_title, propu, shape_uri=uri )
                if valueshape_uri.startswith( self.baseurl):
                    # this shape references another shape - need to load this!
                    vs_xml = self._get_typeuri_rdf(valueshape_uri)
                    subshape_x = rdfxml.xml_find_element( vs_xml,f'.//rdf:Description[@rdf:about="{valueshape_uri}"]' )
                    if subshape_x is None:
#                        print( f"\n\nSUBSHAPE_X=",ET.tostring( vs_xml),"\n\n" )
                        logger.info( f"SubShape definition for {valueshape_uri} not found!" )
                        continue
                    # recurse to load this shape!
                    self._load_type_from_resource_shape( subshape_x, supershape=(property_title,propu))
                else:
                    logger.info( f"SKIPPED external shape {valueshape_uri=}" )
            else:
                logger.debug( f"{valueshape_uri=}" )
#                if valueshape_uri is None or not valueshape_uri.startswith( self.baseurl):
#                    if not valueshape_uri:
#                        valueshape_uri = f"{uri}#{propnodeid}"
#                    if not self.is_known_property_uri( valueshape_uri ):
#                        # define it
#                        self.register_property( property_title, valueshape_uri, shape_uri=uri )
#                        continue
#                    else:
#                        logger.debug( f"ALREADY KNOWN" )
#                        continue

                pd_u = rdfxml.xmlrdf_get_resource_uri( real_propel, 'oslc:propertyDefinition' )

                # In case of repeated identical property titles on a shape, let's create an alternative name that can (perhaps) be used to disambiguate
                # (at least these don't have duplicates AFAICT)
                altname  = pd_u[pd_u.rfind("/")+1:]
                if '#' in altname:
                    altname  = pd_u[pd_u.rfind("#")+1:]

                if pd_u is not None:
                    if not pd_u.startswith( self.baseurl ):
                        self.register_property(property_title,pd_u, shape_uri=uri, altname=altname)
                        logger.debug( f"+++++++Skipping non-local Property Definition {pd_u}" )
                        continue
                else:
                    logger.debug( f"~~~~~~~Ignoring non-local Property Definition {pd_u}" )

                if self.is_known_property_uri( pd_u,shape_uri=uri,raiseifnotfound=False ):
                    logger.debug( f"ALREADY KNOWN2" )
                    continue
                logger.info( f"Defining property {title}.{property_title} {altname=} {pd_u=} +++++++++++++++++++++++++++++++++++++++" )
                self.register_property(property_title,pd_u, shape_uri=uri, altname=altname)
                # check for any allowed value
                allowedvalueu = rdfxml.xmlrdf_get_resource_uri(real_propel, ".//oslc:allowedValue" )
                if allowedvalueu is not None:
                    logger.info( "FOUND ENUM" )
                    # this has enumerations - find them and record them
                    # retrieve each definition
                    nvals = 0
                    for allowedvaluex in rdfxml.xml_find_elements( real_propel,'.//oslc:allowedValue'):
                        allowedvalueu = rdfxml.xmlrdf_get_resource_uri(allowedvaluex )

                        thisenumx = rdfxml.xml_find_element( shapedef,f'.//rdf:Description[@rdf:about="{allowedvalueu}"]' )

                        enum_uri = allowedvalueu
                        logger.info( f"{enum_uri=}" )
                        nvals += 1
                        if not self.is_known_enum_uri( enum_uri ):
                            # retrieve it and save the enumeration name and uri in types cache
                            enum_value_name = rdfxml.xmlrdf_get_resource_text(thisenumx, 'rdfs:label')
                            enum_id = enum_value_name
                            if enum_value_name is None:
                                logger.debug( "enum xml=",ET.tostring(thisenumx) )
                                logger.debug( f"{enum_id=} no name" )
                                raise Exception( "Enum name not present!" )

                            logger.info( f"defining enum value {enum_value_name=} {enum_id=} {enum_uri=}" )
                            self.register_enum( enum_value_name, enum_uri, property_uri=pd_u, id=None )

                    if nvals==0:
                        raise Exception( f"Enumeration {valueshape_uri} with no values loaded" )
        logger.debug( "Finished loading typesystem")
        return n

    # this is a local cache only for the typesystem retrieval
    # the cache deliberately strips off the fragment because it's irrelevant for the GET
    # this makes a lot of the URIs repeats so they are cached
    def _get_typeuri_rdf(self,uri):
        # strip off the fragment as it does nothing
        realuri = uri.rsplit( '#',1 )[0]
        if realuri not in self._gettypecache.keys():
            self._gettypecache[realuri] = self.execute_get_rdf_xml(uri,intent="Retrieve type definition" )
        return self._gettypecache[realuri]

    # given a type URI, return its name
    def resolve_uri_to_name(self, uri, prefer_same_as=True, dontpreferhttprdfrui=True):
        logger.info( f"resolve_uri_to_name {uri=}" )
        if not uri:
            result = None
            return result
        if not uri.startswith('http://') or not uri.startswith('https://'):
        # try to remove prefix
            uri1 = rdfxml.tag_to_uri(uri,noexception=True)
            logger.debug(f"Trying to remove prefix {uri=} {uri1=}")
            if uri1 is None:
                return uri
            if uri1 != uri:
                logger.debug( f"Changed {uri} to {uri1}" )
            else:
                logger.debug( f"NOT Changed {uri} to {uri1}" )
            # use the transformed URI
            uri = uri1
        if not uri.startswith(self.reluri()):
            if self.server.jts.is_user_uri(uri):
                result = self.server.jts.user_uritoname_resolver(uri)
                logger.debug(f"returning user")
                return result
            uri1 = rdfxml.uri_to_prefixed_tag(uri,noexception=True)
            logger.debug(f"No app base URL {self.reluri()=} {uri=} {uri1=}")
            return uri1
        elif not self.is_known_uri(uri):
            if self.server.jts.is_user_uri(uri):
                result = self.server.jts.user_uritoname_resolver(uri)
            else:
                if uri.startswith( "http://" ) or uri.startswith( "https://" ):
                    uri1 = rdfxml.uri_to_prefixed_tag(uri)
                    logger.debug( f"Returning the raw URI {uri} so changed it to prefixed {uri1}" )
                    uri = uri1
                result = uri
            # ensure the result is in the types cache, in case it recurrs the result can be pulled from the cache
            self.register_name(result,uri)
        else:
            result = self.get_uri_name(uri)
        logger.info( f"Result {result=}" )
        return result
        
    # for OSLC query, given an attribute (property) name return its type URI
    # the context is the shape definition - can be None, needed to be specified ultimately by the user when property names aren't unique
    def resolve_property_name_to_uri(self, name, shapeuri=None, exception_if_not_found=True):
        logger.info( f"resolve_property_name_to_uri {name=} {shapeuri=}" )
        result = self.get_property_uri(name,shape_uri=shapeuri)
        logger.info( f"resolve_property_name_to_uri {name=} {shapeuri=} {result=}" )
        return result



##
## © Copyright 2022- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##

#
# convert elmclient log file into a sequence diagram in a html page
# reuses https://bramp.github.io/js-sequence-diagrams/
#

import argparse
import base64
import glob
import html
import io
import json
import os.path
import re
import urllib.parse
import webbrowser
import xml.etree.ElementTree as ET

import jinja2

JINJA_TEMPLATE = """
<!doctype html>
<html>
<head>
    <link rel="stylesheet" type="text/css" href=""https://bramp.github.io/js-sequence-diagrams/css/sequence-diagram-min.css" media="screen" />
</head>
<body onload='showseq();'>
<p>Sequence diagram from emlclient log {{ logfilename }} - wait a few seconds for the diagram to appear...</p>

<div id="diagram"></div>

<script src="https://cdn.jsdelivr.net/npm/underscore@1.13.2/underscore-umd-min.js" ></script>
<script src="https://cdnjs.cloudflare.com/ajax/libs/webfont/1.3.0/webfont.js"></script>
<script src="https://cdnjs.cloudflare.com/ajax/libs/snap.svg/0.5.1/snap.svg-min.js" ></script>
<script src="https://bramp.github.io/js-sequence-diagrams/js/sequence-diagram-min.js" ></script>

<script type="text/javascript"> 
function showseq(){
//  console.log("Hello2");
  var seqdiv =  document.getElementById("sequence");
  seqtext = seqdiv.textContent
  var d = Diagram.parse(seqtext);
  var options = {theme: 'simple'};
  d.drawSVG('diagram', options);
}
</script>
<p>Created using <a href='https://github.com/bramp/js-sequence-diagrams'>js-sequence-diagrams</a> see examples <a href='https://bramp.github.io/js-sequence-diagrams/'>here</a><p>

<div id="sequence" style="display: none;">{{ seq }}</div>

</body>
</html>
"""

# these are case-sensitive
reqshowheaders=[
    'Accept',
    'Configuration-Context',
    'Content-Type',
    'Cookie',
    'DoorsRP-Request-Type',
    'If-Match',
    'net.jazz.jfs.owning-context',
    'OSLC-Core-Version',
    'Referer',
    'vvc.configuration',
]

# these are case-sensitive
respshowheaders=[
    'ETag',
    'Content-Type',
    'Content-Length',
    'Location',
    'Set-Cookie',
    'WWW-Authenticate',
    'X-jazz-web-oauth-url',
    'X-JSA-AUTHORIZATION-REDIRECT',
    'X-JSA-AUTHORIZATION-URL',
    'X-JSA-LOGIN-REQUIRED',
]

# these are case-sensitive
ignore_mimetypes=[
    "application/javascript",
    "application/octet-stream",
    "application/x-javascript",
    "image/gif",
    "image/jpeg",
    "image/png",
    "image/svg+xml",
    "image/x-icon",
    "text/css",
    "text/javascript",
]


def sanitise( s, *, width=60, wrap=True, clip=None, maxlines=0 ):
#    print( f"Sanitising {s=}" )
    results = []
    s0 = s
    s=s.replace('\n',"\\n")
    s=s.replace('\r',"")
    s=s.replace('\\r',"") # to tidy up JSON
    slines = s.split("\\n")
#    print( f"{len(slines)=}" )
    if wrap:
        lines = 0
        while s[:width] and (maxlines==0 or lines<maxlines):
            lines += 1
            results.append( s[:width] )
            s = s[width:]
    elif clip:
        if len(s)>clip:
            results.append( s[:clip]+"..." )
        else:
            results.append( s[:clip] )
    else:
        if maxlines>0:
#            print( f"{maxlines=}" )
#            print( f"{len(slines)=}" )
            if len(slines)>maxlines:
                slines = slines[:maxlines]+['...']
            else:
                slines = slines[:maxlines]
            
        results.extend(slines)
    return html.escape("\\n".join(results))

def addheaders(events,msgheaders,headerstoshow,direction):
    msgs = []
    for header,value in sorted(msgheaders):
        if header in headerstoshow:
            if header == "Cookie":
                cookies = value.split("; ")
                msgs.extend( [f"{sanitise('Cookie: '+c, wrap=False, clip=50)}" for c in cookies] )
            else:
                msgs.append( f"{sanitise(header+': '+value, wrap=False, clip=50)}" )
    if msgs:
        text = '\\n'.join(msgs)
        events.append( f"Note {direction}: {text}" )
    return
    
# parse a message from the logfile into a reqat/response tuple
def decodemessage(msg):
    request = {'intent':None,'method': None, 'url': None, 'headers':[], 'body':None}
    response = {'status': None,'headers':[], 'body':None}
    parts = re.search( r"^(?:INTENT: (.*?)\n\n)?(GET|PUT|POST|HEAD|DELETE|POST) +(\S+)\n( .*?)\n\n(?::+?=\n(.*?)\n-+?=\n)?.*?\n\nResponse: (\d+)\n( .*?)\n\n(?::+?@\n(.*?)\n-+?@\n)?",msg, flags=re.DOTALL+re.MULTILINE )
    request['intent'] = parts.group(1)
    request['method'] = parts.group(2)
    request['url'] = parts.group(3)
    for hdrline in parts.group(4).split( "\n" ):
        hsplit = hdrline.strip().split(": ",1)
        hsplit.append("")
        hdr, value = hsplit[0],hsplit[1]
        request['headers'].append( (hdr,value) )
    request['body'] = parts.group(5).strip() if parts.group(5) else None
    response['status'] = parts.group(6)
    for hdrline in parts.group(7).split( "\n" ):
        hsplit = hdrline.strip().split(": ",1)
        hsplit.append("")
        hdr, value = hsplit[0],hsplit[1]
        response['headers'].append( (hdr,value) )
    response['body'] = parts.group(8).strip() if parts.group(8) else None
    return (request,response)
    
def findheader(hdrname,requestorresponse,notfoundreturn=None):
    for h,v in requestorresponse['headers']:
        if hdrname == h:
#            print( f"FOUND {h=} {v=}" )
            return v
    print( f"Not found {hdrname}" )
    return notfoundreturn

def contentshouldbeshown(requestorresponse):
    # find Content-Type header
    contenttype = findheader("Content-Type",requestorresponse)
    # check it
    if contenttype:
        if contenttype.startswith("application"):
            return True
    return False
    
def isinteger(s):
    try:
        return int(s)>=0
    except:
        pass
    return False
    
#
# Input options:
# Either params are: an ordinal of the receness of the log (0 is latest, 1 is second, etc.)
# Or a filename
#
# output filename
#
def main():
    parser = argparse.ArgumentParser(description="Convert Log file to html page of sequence diagram using https://github.com/bramp/js-sequence-diagrams")
    
    parser.add_argument( "logfile", nargs='*', default=[], help="The names or ordinals of the log file - use 0 (or nothing) to convert the most recent log, 1 for the second newest, etc. OR specify the explicit filenames" )
    parser.add_argument('-o', '--outputfolder', default=None, help=f'Folder for output files')

    args = parser.parse_args()

    # if no logfile specified, assert at least the most recent logfile
    args.logfile = args.logfile or [0]
    
    for logfile in args.logfile:
        if os.path.isfile(logfile):
            pass
        else:
            # see if it's an integer
            if isinteger(logfile) and int(logfile)>=0:
                logfileno = int(logfile)
                alllogs = sorted(glob.glob("logs/*.log"), key=os.path.getctime,reverse=True)
                if not alllogs:
                    raise Exception( "No logs in logs folder!" )
                if logfileno>len(alllogs):
                    raise Exception( f"You asked for index {logfile} but there are only {len(alllogs)} logs!" )
                logfile = alllogs[logfileno]
                print( f"Selected log {logfileno} file {logfile}" )
            else:
                raise Exception( f"{logfile} is neaither a real file nor an integer >0 to indicate the index from most recent (0) to older" )
                
        outname = os.path.splitext(logfile)[0]+".html"
        
        if args.outputfolder:
            outfile = os.path.join(args.outputfolder, outname )
        else:
            outfile = os.path.abspath(outname)
            
        print( f"Processing {logfile} to {outname}" )
        
        log = open(logfile,'rt').read()
        
        # find all the messages using the markers put in by httpops.py
        msgs = re.findall(r"^>>>>>>>>+!\n(.*?\n)<<<<<<+!",log, flags=re.DOTALL + re.MULTILINE)
        
        if not msgs:
            raise Exception( "No messages in {logfile} - you probably need to add option -L TRACE" )
        
        events = []
        host = "elm"

        # find the commandline to put in the title of the sequence diagram
        if cmdline := re.search( r"COMMANDLINE: (.*)", log ):
            cmd = cmdline.group(1)
            events.append( f"Title: Command: {cmd}" )
            
        events.append( f"participant me" )
        events.append( f"participant {host}" )
        
        for i,msg in enumerate(msgs):
            request,response = decodemessage(msg)
            #  if INTENT: then add a note for it

            line = i

            reqparts = urllib.parse.urlparse(request['url'])
            
            path = reqparts.path
            if reqparts.query:
                path += "?"+reqparts.query

            if request['intent']:
                events.append( f"Note left of me: INTENT: {request['intent']}" )
                
            events.append( f"me->{host}: {request['method']} {sanitise(path,wrap=True)}" )
            
            addheaders(events,request['headers'],reqshowheaders,"left of me" )
            
            if request['body']:
                if contentshouldbeshown(request):
                    reqbody = sanitise(request['body'], wrap=True)
                else:
                    reqbody = sanitise(request['body'], wrap=False, maxlines=5)
                    
                if reqbody:
                    events.append( f"Note left of me: {reqbody}" )
                    
            events.append( f"{host}-->me: {response['status']}" )
            
            addheaders(events,response['headers'],respshowheaders,f"right of {host}" )
            
            if response['body']:
                if contentshouldbeshown(response):
                    respbody = sanitise(response['body'], wrap=False)
                else:
                    respbody = sanitise(response['body'], wrap=False, maxlines=5)
#                    respbody = "Content type not shown..."
               
                events.append( f"Note right of {host}: {respbody}" )

        j2_template = jinja2.Template(JINJA_TEMPLATE)
        data = {
            'seq': "\n".join(events),
            'logfilename': os.path.basename(logfile),
        }
        
        # save to file
        open(outfile,"wt").write( j2_template.render(data) )
        
        # display the report
        url = f'file://{os.path.abspath(outfile)}'
        webbrowser.open(url, new=2)  # open in new tab

if __name__=="__main__":
    main()
    


##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


import pathlib
from setuptools import find_packages,setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="elmclient",
    version="0.6.1",
    description="Python client for ELM with examples of OSLC Query, ReqIF import/export, Reportable REST, and more",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/ELM-Python-Client",
    author="Ian Barnard",
#    author_email="ian.barnard@uk.ibm.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["elmclient", "elmclient.examples", "elmclient.examples.basic","elmclient.tests"],
    include_package_data=True,
    install_requires=['CacheControl==0.12.6','anytree==2.8.0',"colorama==0.4.4","cryptography==3.4.4",'lark_parser==0.12.0','lockfile==0.12.2','lxml==4.7.1',"openpyxl == 3.0.9","python-dateutil==2.8.2", "pytz==2021.3", "requests==2.24.0","requests_toolbelt==0.9.1",'tqdm==4.56.2','urllib3==1.25.11'],
    entry_points={
        "console_scripts": [
            "oslcquery=elmclient.examples.oslcquery:main",
            "batchquery=elmclient.examples.batchquery:main",
            "represt=elmclient.examples.represt:main",
            "reqif_io=elmclient.examples.reqif_io:main",
            "log2seq=elmclient.examples.log2seq:main",
        ]
    },
)



##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##

# example of accessing the module structure API https://jazz.net/wiki/bin/view/Main/DNGModuleAPI
# also see https://jazz.net/wiki/bin/view/Main/DNGModuleApiOverview
# prints the module content with indenting corresponding to headings and calculated section number
# NOTE NOTE NOTE the section number calculation has not been fully verified/checked - it seems to work after superficial inspection

# provide on the commandline the id of an artifact in the same component and a new binding for it will be created in the structure - location hardcoded!

import csv
import logging
import os.path
import sys
import time

import lxml.etree as ET
import tqdm

import elmclient.server as elmserver
import elmclient.utils as utils
import elmclient.rdfxml as rdfxml

# setup logging - see levels in utils.py
#loglevel = "INFO,INFO"
loglevel = "TRACE,OFF"
levels = [utils.loglevels.get(l,-1) for l in loglevel.split(",",1)]
if len(levels)<2:
    # assert console logging level OFF if not provided
    levels.append(None)
if -1 in levels:
    raise Exception( f'Logging level {loglevel} not valid - should be comma-separated one or two values from DEBUG, INFO, WARNING, ERROR, CRITICAL, OFF' )
utils.setup_logging( filelevel=levels[0], consolelevel=levels[1] )

logger = logging.getLogger(__name__)

utils.log_commandline( os.path.basename(sys.argv[0]) )

jazzhost = 'https://jazz.ibm.com:9443'
    
username = 'ibm'
password = 'ibm'

jtscontext = 'jts'
rmcontext  = 'rm'

proj = "rm_optout_p1"
comp = proj
conf =  f"{comp} Initial Stream"
mod = "AMR Stakeholder Requirements Specification"

# choose a format to retrieve the structure
format = "JSON"
#format  ="RDFXML"

print( f"Retreiving structure in {format}" )

# caching control
# 0=fully cached (but code below specifies queries aren't cached) - if you need to clear the cache, delet efolder .web_cache
# 1=clear cache initially then continue with cache enabled
# 2=clear cache and disable caching
caching = 2

##################################################################################
# converts heading level into a string that looks like "section" when you add that column into a module
# not exhaustively tested but seems to work :-)
# NOTE NOTE NOTE the section number calculation has not been fully verified/checked - it seems to work after superficial inspection
def getsectionnumber( headinglevel):
    hs = []
    for hn,tn in headinglevel:
        if tn:
            hs.append( f"{hn}-{tn}" )
        else:
            hs.append( f"{hn}" )
    h = ".".join(hs)
    return h


##################################################################################
# iterator to walk the XML hierarchy in a way that lets us track entry/exit from a nesting
def iterwalk(root, events=None, tags=None):
    """Incrementally walks XML structure (like iterparse but for an existing ElementTree structure)
    Returns an iterator providing (event, elem) pairs.
    Events are start and end
    events is a list of events to emit - defaults to ["start","end"]
    tags is a single tag or a list of tags to emit events for - if None or empty list then events are generated for all tags
    """
    # each stack entry consists of a list of the xml element and a second entry initially None
    # if the second entry is None a start is emitted and all children of current element are put into the second entry
    # if the second entry is a non-empty list the first item in it is popped and then a new stack entry is created
    # once the second entry is an empty list, and end is generated and then stack is popped
    stack = [[root,list(root)]]
#    tags = tags if type(tags) == list else tags or []
    tags = [] if tags is None else tags if type(tags) == list else [tags]
    events = events or ["start","end"]
    def iterator():
        while stack:
            elnow,children = stack[-1]
            if children is None:
                # this is the start of elnow so emit a start and put its children into the stack entry
                if ( not tags or elnow.tag in tags ) and "start" in events:
                    yield ("start",elnow)
                # put the children into the top stack entry
                stack[-1][1] = list(elnow)
            elif len(children)>0:
                # do a child and remove it
                thischild = children.pop(0)
                # and now create a new stack entry for this child
                stack.append([thischild,None])                
            else:
                # finished these children - emit the end
                if len(stack)>1 and ( not tags or elnow.tag in tags ) and "end" in events:
                    yield ("end",elnow)
                stack.pop()
    return iterator

def iterwalk1(root, events=None, tags=None):
    """Recuirsive version - Incrementally walks XML structure (like iterparse but for an existing ElementTree structure)
    Returns an iterator providing (event, elem) pairs.
    Events are start and end
    events is a list of events to emit - defaults to ["start","end"]
    tags is a single tag or a list of tags to emit events for - if None or empty list then events are generated for all tags
    """
    tags = [] if tags is None else tags if type(tags) == list else [tags]
    events = events or ["start","end"]
    
    def recursiveiterator(el,suppressyield=False):
        if not suppressyield and ( not tags or el.tag in tags ) and "start" in events:
            yield ("start",el)
        for child in list(el):
            yield from recursiveiterator(child)
        if not suppressyield and  ( not tags or el.tag in tags ) and "end" in events:
            yield ("end",el)
            
    def iterator():
        yield from recursiveiterator( root, suppressyield=True )
        
    return iterator


##################################################################################
if __name__=="__main__":

    # create our "server" which is how we connect to DOORS Next
    # first enable the proxy so if a proxy is running it can monitor the communication with server (this is ignored if proxy isn't running)
    elmserver.setupproxy(jazzhost,proxyport=8888)
    theserver = elmserver.JazzTeamServer(jazzhost, username, password, verifysslcerts=False, jtsappstring=f"jts:{jtscontext}", appstring='rm', cachingcontrol=caching)

    # create the RM application interface
    dnapp = theserver.find_app( f"rm:{rmcontext}", ok_to_create=True )

    # open the project
    p = dnapp.find_project(proj)

    # find the component
    c = p.find_local_component(comp)
    comp_u = c.project_uri
    print( f"{comp_u=}" )

    # select the configuration
    config_u = c.get_local_config(conf)
    print( f"{config_u=}" )
    c.set_local_config(config_u)

    # find the module - using OSLC Query

    # get the query capability base URL for requirements
    qcbase = c.get_query_capability_uri("oslc_rm:Requirement")

    # query for a title and for format=module
    modules = c.execute_oslc_query(
        qcbase,
        whereterms=[['dcterms:title','=',f'"{mod}"'], ['rdm_types:ArtifactFormat','=','jazz_rm:Module']],
        select=['*'],
        prefixes={rdfxml.RDF_DEFAULT_PREFIX["dcterms"]:'dcterms',rdfxml.RDF_DEFAULT_PREFIX["rdm_types"]:'rdm_types',rdfxml.RDF_DEFAULT_PREFIX["jazz_rm"]:'jazz_rm'} # note this is referest - url to prefix
        )
        
    if len(modules)==0:
        raise Exception( f"No module '{mod}' with that name in project {proj} component {comp} configuration {conf}" )
    elif len(modules)>1:
        for k,v in modules.items():
            print( f'{k} {v.get("dcterms:title","")}' )
        raise Exception( "More than one module with that name in project {proj} component {comp} configuraition {conf}" )

    # we've found the module, it's the only entry in the modules dictionary, keyed by URL
    themodule_u = list(modules.keys())[0]
    print( f"{themodule_u=}" )

    mod_x = c.execute_get_rdf_xml(themodule_u, cacheable=False,  headers={'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'Referer': None, 'OSLC-Core-Version': None, 'Configuration-Context': None}, intent="Retrieve the module RDF-XML to get the structure URI" ) # have to remove the OSLC-Core-Version and Configuration-Context headers, and provide vvc.configuration header

    print( f"{mod_x=}" )

    # report the structure

    structure_u = rdfxml.xmlrdf_get_resource_uri( mod_x, ".//rm_modules:structure" )
    print( f"{structure_u=}" )

    if len(sys.argv)>1:
        # second argument is an ID - it *MUST* be in the same configuration as the module!
        # first find the artifact to insert, using OSLC Query
        # query for a title and for format=module
        toinserts = c.execute_oslc_query(
            qcbase,
            whereterms=[['dcterms:identifier','=',sys.argv[1]]],
            select=['*'],
            prefixes={rdfxml.RDF_DEFAULT_PREFIX["dcterms"]:'dcterms'}
            )
            
        if len(toinserts)==0:
            raise Exception( f"No artifact '{sys.argv[1]}' with that name in project {proj} component {comp} configuration {conf}" )
        elif len(toinserts)>1:
            toinsert_u = None
            for k,v in toinserts.items():
                # find the first one with a rm_nav:parent
                folder = v.get("rm_nav:parent")
                if folder is not None:
                    toinsert_u = k
                    break
            if toinsert_u is None:
                raise Exception( f"No artifact {sys.argv[1]} found with a folder" )
        else:
            toinsert_u = list(toinserts.keys())[0]
        print( f"{toinsert_u=}" )
        
    if format == "RDFXML":
        # retrieve the structure element in RDF-XML
        modstructure_x = c.execute_get_rdf_xml(structure_u, cacheable=False, headers={'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'OSLC-Core-Version': None, 'Configuration-Context': None}, intent="Retrieve module structure (XML)" ) # have to remove the OSLC-Core-Version and Configuration-Context headers, and provide vvc.configuration header

        if len(sys.argv)>1:            
            # toinsert is already prepared
            # get the etag
            response, etag = c.execute_get_rdf_xml(structure_u, cacheable=False, headers={'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'OSLC-Core-Version': None, 'Configuration-Context': None}, return_etag=True, intent="Retrieve module structure (XML)"  ) # have to remove the OSLC-Core-Version and Configuration-Context headers, and provide vvc.configuration header
            print( f"{etag=}" )
            # insert a reference to it in a hardcoded location
            firsthead_x = rdfxml.xml_find_elements(modstructure_x,'rm_modules:Binding/rm_modules:childBindings/rm_modules:Binding/rm_modules:childBindings')[0]
            print( f"{firsthead_x=}" )
#        <rm_modules:childBindings rdf:parseType="Collection">
#            <rm_modules:Binding rdf:about="https://clmwb.com:9444/rdm/resources/MB_2f75a3f310ec42ccadb68e4442617a74">
#                <rm_modules:isHeading rdf:datatype="http://www.w3.org/2001/XMLSchema#boolean">true</j.0:isHeading>
#                <oslc_config:component rdf:resource="https://clmwb.com:9444/rdm/cm/component/_Y5OakLruEeevMtDqCXe--Q"/>
#                <rm_modules:boundArtifact rdf:resource="https://clmwb.com:9444/rdm/resources/CA_3a7b45d1caa540d485ffc9832b802d47"/>
#                <rm_modules:module rdf:resource="https://clmwb.com:9444/rdm/resources/_4sscEb43EeeD0-df1VhHuw"/>
#                <rm_modules:childBindings rdf:parseType="Collection">

#            s1 = ET.SubElement( firsthead_x )

            # add a new childbinding
#                    <rm_modules:Binding rdf:about="https://clmwb.com:9444/rdm/resources/_4sscEb43EeeD0-df1VhHuw/structure#1">
#                        <oslc_config:component rdf:resource="https://clmwb.com:9444/rdm/cm/component/_Y5OakLruEeevMtDqCXe--Q"/>
#                        <rm_modules:boundArtifact rdf:resource="https://clmwb.com:9444/rdm/resources/CA1"/>
#                        <rm_modules:module rdf:resource="https://clmwb.com:9444/rdm/resources/_4sscEb43EeeD0-df1VhHuw"/>
#                        <rm_modules:childBindings rdf:resource="http://www.w3.org/1999/02/22-rdf-syntax-ns#nil"/>
#                    </rm_modules:Binding>
            newbinding = ET.fromstring(
                f"""<rm_modules:Binding
                            xmlns:rdf='{rdfxml.RDF_DEFAULT_PREFIX["rdf"]}'
                            xmlns:oslc_config='{rdfxml.RDF_DEFAULT_PREFIX["oslc_config"]}'
                            xmlns:rm_modules='{rdfxml.RDF_DEFAULT_PREFIX["rm_modules"]}'
                            rdf:about="https://clmwb.com:9444/rdm/resources/_4sscEb43EeeD0-df1VhHuw/structure#1">
                        <oslc_config:component rdf:resource="{comp_u}"/>
                        <rm_modules:boundArtifact rdf:resource="{toinsert_u}"/>
                        <rm_modules:module rdf:resource="{themodule_u}"/>
                        <rm_modules:childBindings rdf:resource="http://www.w3.org/1999/02/22-rdf-syntax-ns#nil"/>
                    </rm_modules:Binding>
                """
                )
            print( f"{newbinding=}" )
            firsthead_x.append(newbinding)
            
            # PUT the new structure and wait for it to be saved
            response = c.execute_post_rdf_xml( structure_u, data=modstructure_x, put=True, cacheable=False, headers={'If-Match':etag,'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'OSLC-Core-Version': None, 'Configuration-Context': None}, intent="Update module structure (XML)"  )
            print( f"{response.status_code}" )
            location = response.headers.get('Location')
            if response.status_code == 202 and location is not None:
                # wait for the tracker to finished
                result = c.wait_for_tracker( location, interval=1.0, progressbar=True, msg=f"Updating Structure")
                time.sleep( 0.5 )
                    
            # get the structure again
            modstructure_x = c.execute_get_rdf_xml(structure_u, cacheable=False, headers={'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'OSLC-Core-Version': None, 'Configuration-Context': None}, intent="Retrieve module structure (XML) after update"  ) # have to remove the OSLC-Core-Version and Configuration-Context headers, and provide vvc.configuration header
            
        # find the root binding
        modroot_x = rdfxml.xml_find_element( modstructure_x, './rm_modules:Binding' )

            
        # Now explore the structure for childBinding (which corresponds to nesting) and Binding (which is a binding of an artifact into the module)
        it = iterwalk1(modroot_x, events=["start","end"], tags=[rdfxml.uri_to_tag('rm_modules:childBindings'),rdfxml.uri_to_tag("rm_modules:Binding")] )
        print( f"{it=}" )
        level = 0

        headinglevel = [] # heading level is a list of two-element lists - first is the heading number, second is the non-heading number

        for event,el in it():
            logger.info( f"{event=} {el.tag=} {headinglevel=}" )
            
            # childBinding is an increase in nesting of headings
            if el.tag==rdfxml.uri_to_tag("rm_modules:childBindings"):
                if event == "start":
                    level += 1
                    headinglevel.append([0,0])
                if event == "end":
                    level -= 1
                    headinglevel.pop()
                    
            # Binding is an artifact
            if el.tag==rdfxml.uri_to_tag("rm_modules:Binding"):
                # if it's a heading this increments section differently from non-heading
                isheading = ( rdfxml.xmlrdf_get_resource_text( el, './rm_modules:isHeading' ) == "true" )
                logger.info( f"{isheading=}" )
                
                if event=="start":
                    # retrieve the title of the artifact, only needed in the "start"
                    ba_u = rdfxml.xmlrdf_get_resource_uri( el, './rm_modules:boundArtifact' )
                    if ba_u and ba_u.startswith( c.app.baseurl ):
                        req_x = c.execute_get_rdf_xml( ba_u, intent="Retrieve artifact detail to get title and identifier"  )
                        summary = rdfxml.xmlrdf_get_resource_text( req_x,'.//dcterms:title')
                        id = rdfxml.xmlrdf_get_resource_text( req_x,'.//dcterms:identifier')
                    else:
                        summary = "TOP LEVEL"
                        id = "-"
                        raise Exception( f"Unexpected: No or invalid artifact URI {ba_u}" )
                    if isheading:
                        # increment the heading number and reset the sub-number
                        headinglevel[-1][0] += 1
                        headinglevel[-1][1] = 0
                    else:
                        # increment the sub-number
                        headinglevel[-1][1] += 1
                    # report the current item
                    print( f"{id}{'    '*level}", end="" )
                    if True or isheading:
                        # NOTE NOTE NOTE the section number calculation has not been fully verified/checked - it seems to work after superficial inspection
                        h = getsectionnumber(headinglevel)
                        print( f"{h}", end="" )
                    print( f"  {summary}" )

    elif format == "JSON":
        # retrieve the structure element in RDF-XML
        modstructure_j = c.execute_get_json(structure_u, cacheable=False, headers={'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'OSLC-Core-Version': None, 'Configuration-Context': None}, intent="Retrieve module structure (JSON)"  ) # have to remove the OSLC-Core-Version and Configuration-Context headers, and provide vvc.configuration header

        if len(sys.argv)>1:     
            # Following code harcodes location where new binding is inserted, so need the root to be the first element in the structure list
            if not modstructure_j[0]["isStructureRoot"]:
                raise Exception( "Root not at start of structure!" )
            # toinsert is already prepared
            # get the etag
            response,etag = c.execute_get_rdf_xml(
                structure_u,
                cacheable=False,
                headers={
                    'vvc.configuration': config_u,
                    'DoorsRP-Request-Type':'public 2.0',
                    'OSLC-Core-Version': None,
                    'Configuration-Context': None
                },
                return_etag=True,
                intent="Retrieve module structure (JSON)"
            )
            # have to remove the OSLC-Core-Version and Configuration-Context headers, and provide vvc.configuration header
            print( f"{etag=}" )
            # insert a reference to it in a hardcoded location
            # this is a very blunt method - a tidier method would be to find the section and insert at that location
#{
#  "uri" : "https://jazz.ibm.com:9443/rm/resources/BI__yq5s0B6Eeuh3Iiax2L3Ow",
#  "type" : "dng_module:Binding",
#  "component" : "https://jazz.ibm.com:9443/rm/cm/component/__J_JEEB6Eeuh3Iiax2L3Ow",
#  "isHeading" : false,
#  "module" : "https://jazz.ibm.com:9443/rm/resources/MD__y90_0B6Eeuh3Iiax2L3Ow",
#  "boundArtifact" : "https://jazz.ibm.com:9443/rm/resources/TX__2GBIkB6Eeuh3Iiax2L3Ow",
#  "childBindings" : [ ]
#}          
            tempbinding_u = "https://clmwb.com:9444/rdm/resources/_4sscEb43EeeD0-df1VhHuw/structure#1"
            newbinding = {
                    "uri": tempbinding_u,
                    "component": comp_u,
                    "type": "dng_module:Binding",
                    "module": themodule_u,
                    "boundArtifact": toinsert_u,
                    "childBindings": []
                }
            # this assumes the root is the first entry!
            modstructure_j[0]["childBindings"].append(tempbinding_u)
            modstructure_j.append(newbinding)
            # PUT the new structure and wait for it to be saved
            response = c.execute_post_json( structure_u, data=modstructure_j, put=True, cacheable=False, headers={'If-Match':etag,'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'OSLC-Core-Version': None, 'Configuration-Context': None}, intent="Update the module structure (JSON)"  )
            print( f"{response.status_code}" )
            location = response.headers.get('Location')
            if response.status_code == 202 and location is not None:
                # wait for the tracker to finished
                result = c.wait_for_tracker( location, interval=1.0, progressbar=True, msg=f"Updating Structure")
                time.sleep( 0.5 )
            
            # get the structure again afer the update
            modstructure_j = c.execute_get_json(structure_u, cacheable=False, headers={'vvc.configuration': config_u,'DoorsRP-Request-Type':'public 2.0', 'OSLC-Core-Version': None, 'Configuration-Context': None}, intent="Retrieve module structure (JSON) after update"  ) # have to remove the OSLC-Core-Version and Configuration-Context headers, and provide vvc.configuration header
            sys.exit(0)
        # scan all the entries into a dictionary keyed by the URL
        entries = {}
        for s in modstructure_j:
            entries[s["uri"]] = s

        # find the root binding
        modroot = entries[structure_u]

        # generator recursively walking the structure
        def json_structure_walk(rooturi,entries):
            """
            Module structure recursive iterator
            """
            def recursiveiterator(uri,suppressyield=False):
#                print( f"starting {uri}" )
                if not suppressyield:
                    yield ("start",entries[uri])
                yield ("startChildren",entries[uri])
                for child in list(entries[uri]["childBindings"]):
                    yield from recursiveiterator(child)
                yield ("endChildren",entries[uri])
                if not suppressyield:
                    yield ("end",entries[uri])
#                print( f"ending {uri}" )
                    
            def iterator():
                yield from recursiveiterator( rooturi, suppressyield=True )
                
            return iterator        

        # Now explore the structure for childBinding (which corresponds to nesting) and Binding (which is a binding of an artifact into the module)
        it = json_structure_walk(structure_u,entries )
#        print( f"{it=}" )
        level = 0

        headinglevel = [] # heading level is a list of two-element lists - first is the heading number, second is the non-heading number

        for event,el in it():
            
            # childBinding is an increase in nesting of headings
            if event == "startChildren":
                level += 1
                headinglevel.append([0,0])
            if event == "endChildren":
                level -= 1
                headinglevel.pop()
                    
            isheading = el["isHeading"]
            
            if event=="start":
                # retrieve the title of the artifact, only needed in the "start"
                ba_u = el["uri"]
                if ba_u.startswith( c.app.baseurl ):
                    req_x = c.execute_get_rdf_xml( ba_u, intent="Retrieve artifact details to get the title and identifier"  )
                    summary = rdfxml.xmlrdf_get_resource_text( req_x,'.//dcterms:title')
                    id = rdfxml.xmlrdf_get_resource_text( req_x,'.//dcterms:identifier')
                else:
                    summary = "TOP LEVEL"
                    id = "-"
                    raise Exception( f"Unexpected: No or invalid artifact URI {ba_u}" )
                if isheading:
                    # increment the heading number and reset the sub-number
                    headinglevel[-1][0] += 1
                    headinglevel[-1][1] = 0
                else:
                    # increment the sub-number
                    headinglevel[-1][1] += 1
                # report the current item
                print( f"{id}{'    '*level}", end="" )
                if True or isheading:
                    # NOTE NOTE NOTE the section number calculation has not been fully verified/checked - it seems to work after superficial inspection
                    h = getsectionnumber(headinglevel)
                    print( f"{h}", end="" )
                print( f"  {summary}" )

    else:
        raise Exception( f"Invalid format {format}" )
        